// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#ifndef BITSET_H
#define BITSET_H

#include <iostream>

namespace bbq {

/// This class provides one possible way of handling sets, which is a
/// major part of the algorithms underlying \c bbq. This class
/// particularly supports unweighted sets whose cardinality is limited
/// to 32 elements; it supports neither weighing nor multiset
/// capabilities, but is very memory efficient. In its current state,
/// it is only suitable for the ancient \c footprint_detector rather
/// than the featuristic \c weighted_footprint_detector.
class uint_bitset 
{
private:
	
	unsigned int bits;

public:

	// k corresponds to the number of bits and hence the number of
	// sequence fragments in footprint_detector.  In uint64_bitset,
	// this variable only exists for formal reasons and does not have
	// an influence on the number of bits, which is fixed to
	// 8*sizeof(unsigned int), which is usually 32.
	static int k;
	
	uint_bitset()
		{
			bits = 0;
		}

	/// Test whether \c this is a subset of another set.
	bool subseteq(const uint_bitset&);
	
	/// We obtain a canonical order on subsets of [1:k] by considering
	/// such a set as a number between 0 and 2^k-1. This order is
	/// needed for inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator<(const uint_bitset&);

	/// Two sets are considered equivalent iff they contain the same
	/// elements, not regarding their weights. this is needed for
	/// inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator==(const uint_bitset&);

	// this version of subseteq checks only up to the J-th bit whether
	// *this is a subset of the set passed as an argument.
	bool subseteq(const uint_bitset&, const unsigned int& J);
	
	/// returns the number of elements contained in the multiset
	int cardinality();

	friend std::ostream& operator<<(std::ostream& o,const uint_bitset&);
	
	/// standard set intersection
	void intersect(const uint_bitset& b);

	/// add one element j to the multiset. The weight is (currently)
	/// neglected.
	void set(const int& j);

	/// remove ALL elements j from the multiset. Can actually be
	/// ignored for multisets, since the remove job is taken by
	/// decrease.
	void remove(const int& j);
	
};

/// This class provides one possible way of handling sets, which is a
/// major part of the algorithms underlying \c bbq. This class
/// particularly supports unweighted sets whose cardinality is limited
/// to 64 elements; it supports neither weighing nor multiset
/// capabilities, but is very memory efficient.
///
/// For a brief documentation of the individual members, we refer to
/// \c uint_bitset.
class uint64_bitset 
{
private:
	
	unsigned long long bits;

public:

	// k corresponds to the number of bits and hence the number of
	// sequence fragments in footprint_detector.  In uint64_bitset,
	// this variable only exists for formal reasons and does not have
	// an influence on the number of bits, which is fixed to
	// 8*sizeof(unsigned long long), which is usually 64.
	static int k;
	
	uint64_bitset()
		{
			bits = 0;
		}
	
	bool subseteq(const uint64_bitset&);
	
	/// We obtain a canonical order on subsets of [1:k] by considering
	/// such a set as a number between 0 and 2^k-1. This order is
	/// needed for inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator<(const uint64_bitset&) const;

	/// Two sets are considered equivalent iff they contain the same
	/// elements, not regarding their weights. this is needed for
	/// inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator==(const uint64_bitset&) const;

	int cardinality();

	friend std::ostream& operator<<(std::ostream& o,const uint64_bitset&);
	
	void intersect(const uint64_bitset& b);

	void set(const int& j);

	void remove(const int& j);
	
};

/// This class provides one possible way of handling sets, which is a
/// major part of the algorithms underlying \c bbq. This class
/// particularly supports unweighted sets whose cardinality is
/// unlimited. It supports neither weighing nor multiset capabilities,
/// but is more memory efficient than the weighted versions.
///
/// For a brief documentation of the individual members, we refer to
/// \c uint_bitset.
class variable_bitset 
{
private:
	
	unsigned int* bits;

public:

	// k corresponds to the number of bits and hence the number of
	// sequence fragments in footprint_detector
	static int k;

	/// copy constructor
	variable_bitset(const variable_bitset&);

	/// assignment operator
	variable_bitset& operator=(const variable_bitset& vb);
	
	/// default constructor
	variable_bitset();

	/// destrcutor
	~variable_bitset()
		{
			delete[] bits;
		}

	bool subseteq(const variable_bitset&) const;
	
	/// returns the number of elements contained in the set
	int cardinality();

	friend std::ostream& operator<<(std::ostream& o,const variable_bitset&);
	
	void intersect(const variable_bitset& b);

	void set(const int& j);

	void remove(const int& j);
	
	/// We obtain a canonical order on subsets of [1:k] by considering
	/// such a set as a number between 0 and 2^k-1. This order is
	/// needed for inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator<(const variable_bitset&);

	/// Two sets are considered equivalent iff they contain the same
	/// elements, not regarding their weights. this is needed for
	/// inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator==(const variable_bitset&);
	
};


} // end namespace bbq

#endif

