// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include <limits.h>
#include <list>
#include <vector>
#include <iostream>
#include "footprint_detector.h"

namespace bbq {
	
footprint_detector::footprint_detector()
{
	k = 0;
	K = 0;
	s = NULL;
	T = NULL;
	n = NULL;
	N = NULL;
}

footprint_detector::footprint_detector(int k_, int K_, char** s_, char** T_, int* n_, int* N_)
{
	k = k_;
	K = K_;
	s = s_;
	T = T_;
	n = n_;
	N = N_;

	bit_set::k = k;
	
}


void footprint_detector::init_arrangement_cells()
{
	
	int i, j, ii, jj;
	
	for (i=0; i<K; i++)
	{
		
		p[i] = new int*[k];
		lambda[i] = new int[k];
		int* nu = new int[k];

		bit_set current_cell_set;

		// count how many intervals of color j are open
		int* open_interval_counter = new int[k];
		int lambda_i = 0;
		
		for (j=0; j<k; j++)
		{
			nu[j] = 0;
			open_interval_counter[j] = 0;
			// compute all lambda[i][j] many occurences of s[j] in T[i].
			get_occurences(i,j);
			lambda_i += lambda[i][j];
		}
		// lambda_i contains the total number of occurences of fragments
		// in T_i. Apparently, 2*lambda_i is an upper bound for the
		// number of cells we may encounter.
		C[i] = new bit_set[2*lambda_i];
		P[i] = new int[2*lambda_i];

		// yes, we're SWEEPING again!
		int sweep_position = 0;
		int occs_done = 0; // how many p[i][j]s have we sweeped over completely so far?

		// open_intervals[j] contains the list of open intervals of
		// color j. These interval ends are ordered accoring to their
		// positions.
		std::vector< std::list<int> > open_intervals(k);
		int total_occs = 0;
		for (j=0; j<k; j++)
		{
			open_intervals[j].clear();
			total_occs += lambda[i][j];
		}
		
		while (occs_done<total_occs)
		{
			// determine the starting position of the next cell.

			// determine the next BEGINNING of an interval (using p[i][j][nu[j]])
			int best_beg_j = 0;
			int best_beg_p = INT_MAX;
			for (j=0; j<k; j++)
			{
				if (nu[j]<lambda[i][j])
				{
					if (p[i][j][nu[j]] < best_beg_p)
					{
						best_beg_p = p[i][j][nu[j]];
						best_beg_j = j;
					}
				}				
			}

			// determine the next END of an interval
			// (which can be done using the first entries of the k open_intervals lists)
			int best_end_j = 0;
			int best_end_p = INT_MAX;
			for (j=0; j<k; j++)
			{
				if (open_intervals[j].size()>0)
				{
					if (*(open_intervals[j].begin()) < best_end_p)
					{
						best_end_p = *(open_intervals[j].begin());
						best_end_j = j;
					}
				}				
			}

			// Which occurs earlier: the next BEGINNING interval or the
			// next ENDING interval?
			if (best_beg_p <= best_end_p)
			{
				sweep_position = best_beg_p;
				nu[best_beg_j]++;
				open_interval_counter[best_beg_j]++;
				// activate the inteval's color in the set of currently active colors
				current_cell_set.set(best_beg_j);
				// mark the interval as being opened by appending it to
				// the intervals_open[j] list.
				open_intervals[best_beg_j].push_back(best_beg_p + L);
			} else {
				sweep_position = best_end_p;
				open_interval_counter[best_end_j]--;
				if (open_interval_counter[best_end_j]<=0)
					// deactivate the inteval's color in the set of currently active colors
					current_cell_set.remove(best_end_j);
				// unmark the interval by removing it from the
				// intervals_open[j] list.
				open_intervals[best_end_j].pop_front();
				occs_done++;
			}
			// What The F*#!ing Hell has gcc been doing with the
			// assignment operator call? Anyway, now it's working...
			C[i][Lambda[i]] = current_cell_set;
			//bit_set ccs = current_cell_set;
			//C[i][Lambda[i]] = bit_set(current_cell_set);
			P[i][Lambda[i]] = sweep_position;
			Lambda[i]++;
		}		
	}
}

bool footprint_detector::increase(int* nu, int I)
{
	int i;
	bool nu_0_full = (nu[0]==Lambda[0]-1);
	for (i=K-1; i>=0; i--)
	{
		++nu[i];
		if (nu[i]==Lambda[i])
			nu[i] = 0;
		else 
			break;
	}
	return (nu[0]==0 && nu_0_full);
}
	
void footprint_detector::compute_intersections()
{
	int i, mu;
	int C_size = 1;
	int* nu = new int[K];

	overall_best_length = 0;
	
	for (i=0; i<K; i++)
	{
		nu[i] = 0;
		C_size *= Lambda[i];
	}

	bool done = false;
	while (!done)
	{
		// compute the intersection...
		bit_set intersection = C[0][nu[0]];
		bool bound = false;
		for (i=1; i<K; i++)
		{
			intersection.intersect(C[i][nu[i]]);
			if (intersection.cardinality()<=overall_best_length)
			{
				// perform a bound-step in the branch-and-bound-algorithm
				done = increase(nu,i);
				bound = true;
				break;
			}
		}

		if (!bound)
		{
			// if we reach this point, we have
			//
			//    intersection.cardinality() > overall_best_length, 
			//
			// where intersection is the intersection of all C[i][nu[i]],
			// for all i in [1:K]. In other words, this means that we
			// have found a new best intersection.
			bit_set saved_intersection = intersection;
			
			overall_best_length = intersection.cardinality();
			best_intersection = saved_intersection;
			for (int ii=0; ii<K; ii++)
				best_nu[ii] = nu[ii];
			// finally increment the counters...
			done = increase(nu,0);
		}
	}
}

void footprint_detector::get_occurences(int i, int j)
{
	
	std::list<int> match_pos;
	int ii, jj, nu;
	
	for (ii = 0; ii<=N[i]-n[j]; ii++)
	{
		bool match = true;
		for (jj = 0; (jj<n[j]) && match; jj++)
			if (s[j][jj]!=T[i][ii+jj])
				match = false;
		if (match)
		{
			match_pos.push_back(ii);
		}
	}
	int LL = match_pos.size();
	lambda[i][j] = LL;
	p[i][j] = new int[LL];

	std::list<int>::iterator m_it = match_pos.begin();
	std::list<int>::iterator m_end = match_pos.end();;

	nu = 0;
	for (; m_it!=m_end; ++m_it, ++nu)
		p[i][j][nu] = *m_it;
	
}

void footprint_detector::run(int L_)
{
	
	int i, j, ii, jj;

	L = L_;
	
	p = new int**[K];
	lambda =  new int*[K];
	Lambda = new int[K];
	P = new int*[K];
	C = new bit_set*[K];

	best_nu = new int[K];
	
	init_arrangement_cells();

	compute_intersections();
	
}

int footprint_detector::get_result()
{
	return overall_best_length;
}

int footprint_detector::print_result()
{
	std::cout<<"Best intersection: "<<best_intersection<<"\n";
	for (int ii=0; ii<K; ii++)
		std::cout<<"L-Occurence in File "<<ii<<" at pos. "<<P[ii][best_nu[ii]]-L<<".\n";			
	std::cout<<std::endl;
	return overall_best_length;
	std::cout<<std::endl;
}

footprint_detector::~footprint_detector()
{
	int i,j,nu;
	if (p!=NULL)
	{
		for (i = 0; i<K; i++)
		{
			for (j = 0; j<k; j++)
				delete[] p[i][j];
			delete[] p[i];
		}
		delete[] p;
		p = NULL;
	}	
	if (C!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] C[i];
		delete[] C;
		C = NULL;
	}
	if (P!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] P[i];
		delete[] P;
		P = NULL;
	}
	if (lambda!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] lambda[i];
		delete[] lambda;
		lambda = NULL;
	}
	if (Lambda!=NULL)
	{
		delete[] Lambda;
		Lambda = NULL;
	}
}

} // end namespace bbq

