/*
  bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/


#include <ctype.h>
#include <iostream>
#include <fstream>
#include <list>
#include <stdio.h>
#include <map>
#include <string>

#include "footprint_detector.h"
#include "weighted_bitsets.h"
#include "weighted_footprint_detector.h"

#define CHUNK_SIZE 4096


/** \mainpage

\c bbq is a powerful command-line tool for discovering clustered
occurences of transcription factor binding sites in genomes. This is
done in a multiple-alignment-like fashion by solving a certain
geometric optimization problem, the so-called best barbeque problem
(explaining the name \c bbq).

The implementation provided here supports many practically relevant
features, such as weighted sets, p-value based weighting schemes,
computing the best \c h solutions instead only the best solution, and
several others. Most of these features are supported by two algorithms
representing two different approaches to solving the given
optimization problem.

The main goal of this implementation was to achieve best possible
running times - together with a demand for numerous features
(and combinations of features), which meant finding the right balance
between an extendable and maintainable design and high performance.
Templates sometimes lessened the pain of finding the right balance -
anyway, anybody who makes it up to this point in the source code
should keep in mind that this implementation inevitably is the result
of a compromise. Whoever modifies this, should bear in mind to keep it
a high-performance implementation.

A few words should be spent on the class design. All routines for
scanning command line parameters and reading input files are contained in
\c main.cpp in the global namespace. The functions in \c main.cpp create a
suitable instance of the template class \c weighted_footprint_detector.

Class \c weighted_footprint_detector implements all crucial
parts of the underlying algorithms, including string matching, computation
of interval arrangements and finding best intersections of cells (based
on algorithm (A1) or (A2)). Class \c weighted_footprint_detector requires
handling (weighted) sets or multisets. For the different versions -
weighted/unweighted and sets/multisets - different classes can be passed as
a template parameter to \c weighted_footprint_detector: \c weighted_variable_bitset
for weighted bitsets of arbitrary cardinality, \c variable_bitset for arbitrary
length unweighted bitsets, \c uint_bitset and \c uint64_bitset for unweighted
bitsets of limited cardinality (i.e., 32 or 64 elements), \c bit_multi_set for
unweighted multisets and \c weighted_bit_multi_set for weighted multisets.

Due to the performance oriented design, the interface for set-handling
classes is somewhat more complex than might be neccessary in a minimum-effort
implementation. In case of implementing your own set-handling class, refer to
the existing ones. Note that generic programming allows for flexible
algorithms essentially without losing efficiency.

All classes are contained in the namespace \c bbq.

A rather historical version of \c bbq - demonstrating the actual idea behind
the algorithms - can be found in the class \c footprint_detector.

*/



using namespace bbq;
	
	
char** T;
char** s;
char** s_names;
int* n;
int* N;
int* m;
int L;
int k;
int K;

std::set<std::string> groups;
std::map< std::string,std::list<std::string> > group_members;

void fill_reverse_segments();

/// Are the parameters passed to the program legal?
bool params_good;

/// number of command-line options
int num_of_opts;

/// number of hits to be computed
int num_of_hits;

/// for -t option; threshold <0. means that only an optimal solution is
/// computed; threshold >=0 means that only all solutions >=threshold
/// are computed.
float threshold;

/// if the threshold was specified by -tc <f> rather than by -t <f>,
/// we only count solutions -- instead of printing every solution --
/// matching the threshold. This feature can be very useful to find a
/// suitable threshold that yields a reasonable number of matches.
bool count_results;

/// consider reverse complemented segments also?
bool reverse;

/// print progress report while algorithm is running? 
bool report_progress;

/// choose algorithm (A1) (default) or (A2)
int algorithm;

/// false if -u option (unweighted) is given.
bool weighted;

/// see \c weighted_footprint_detector::weight_type for a description
/// of different weight types available.
int weight_type;

/// group fragments into classes based on the first four letters
bool grouped;

/// true if multiple occurences within one cluster are to make a
/// difference.
bool use_multisets;

/// true if postscript output (-P <filename>) was specified.
bool ps_output;
std::string ps_file_name;

/// give more verbose output for stupid users
bool stupid_user;

/// see documentation of
/// weighted_footprint_detector::num_of_group_characters
int num_of_group_characters;

/// Use a fast algorithm that only takes into account matches where
/// (set theoretic) difference between the sets involved in the match
/// is bounded by \c delta.
bool bounded_difference;

/// bound for the number of non-matching fragments contined in a
/// match. The bound \c delta is taken into account if \c
/// bounded_difference=\c true (using command line option \c -\c D).
int delta;

/// reads fname (in fasta format) into T[i].
bool read_genome(char* fname, int i)
{
	std::ifstream ifs(fname);
	std::list<char*> chunk_list;
	int last_chunk_length = 0;
	int total_length = 0;
	char* chunk_block;
	// overread first line
	char ch = 0;
	if (ifs.good())
		ifs.get(ch);
	else
		return false;
	if (ch=='>')
	{
		// overread fasta-stuff
		bool go_on = ifs.good();
		while (go_on)
		{
			char ch;
			ifs.get(ch);
			go_on = (ch!='\n' && ifs.good());
		}
	} else {
		ifs.unget();
	}
	bool final_chunk = false;
	while (!final_chunk)
	{
		char* chunk_tmp = chunk_block;
		chunk_block = new char[CHUNK_SIZE];
		int chunk_length = ifs.readsome(chunk_block,CHUNK_SIZE);
		if (chunk_length!=0)
		{
			chunk_list.push_back(chunk_block);
			last_chunk_length = chunk_length;
		} else {
			final_chunk = true;
			chunk_block = chunk_tmp;
		}
		total_length += chunk_length;
	}
	if (last_chunk_length<0 || total_length<0)
		return false;
	
	T[i] = new char[total_length];
	char* T_i = T[i];
	std::list<char*>::iterator c_it = chunk_list.begin();
	int chs = chunk_list.size()-1;
	int pos = 0;
	int ii,j;
	for (ii=0; ii<chs; ii++, ++c_it)
	{
		char* CC = *c_it;
		for (j=0; j<CHUNK_SIZE; j++)
		{
			if (CC[j]>32)
			{
				T_i[pos] = CC[j];
				pos++;
			}
		}
		delete[] CC;
	}
	for (j=0; j<last_chunk_length; j++)
			if (chunk_block[j]>32)
			{
				T_i[pos] = chunk_block[j];
				pos++;
			}
	delete[] chunk_block;

	N[i] = pos;
	int N_i = pos;
	
	int iota;
	for (iota=0; iota<N_i; iota++)
		T_i[iota] = toupper(T_i[iota]);
	
	return (total_length != 0);
}


/// read over all useless characters such as spaces, tabs and line
/// breaks
void skip(std::ifstream& ifs)
{
	if (!ifs.good())
		return;
	char ch = 0;
	ifs.get(ch);
	while (ifs.good() && ch<=32)
		ifs.get(ch);
	if (ifs.good())
		ifs.unget();
}

bool read_parameters(int argc, char** argv)
{
	if (argc<2)
		return false;
	
	weighted = false;
	weight_type = 1;
	params_good = true;
	stupid_user = false;
	report_progress = false;
	reverse = false;
	algorithm = 1;
	use_multisets = false;
	ps_output = false;
	bounded_difference = false;
	delta = 0;
	num_of_opts = 0;
	num_of_hits = 1;
	// threshold <0. means that only an optimal solution is computed; 
	// threshold >=0 means that only all solutions >=threshold are computed.
	threshold = -1.;
	count_results = false;
	num_of_group_characters = 5;
	int i;
	for (i=1; i<argc && argv[i][0]=='-'; i++)
	{
		num_of_opts++;
		switch (argv[i][1])
		{
			case 'S' : {
				stupid_user = true;
				break;
			}
			case 'u' : {
				weighted = false;
				weight_type = 0;
				break;
			}
			case 'r' : {
				reverse = true;
				break;
			}
			case 'p' : {
				report_progress = true;
				break;
			}
			case 'g' : {
				grouped = true;
				if (i+1<argc && argv[i+1][0]!='-')
				{
					// read num_of_group_characters
					num_of_group_characters = atoi(argv[i+1]);
					i++;
					num_of_opts++;
				} 
				break;
			}
			case 'D' : {
				bounded_difference = true;
				if (i+1<argc && argv[i+1][0]!='-')
				{
					// read num_of_group_characters
					delta = atoi(argv[i+1]);
					i++;
					num_of_opts++;
				} 
				break;
			}
			case 'A' : {
				if (argv[i][2]=='1')
					algorithm = 1;
				else if (argv[i][2]=='2')
					algorithm = 2;
				break;
			}			case 'h' : {
				if (i+1<argc && argv[i+1][0]!='-')
				{
					// read num_of_hits
					num_of_hits = atoi(argv[i+1]);
					i++;
					num_of_opts++;
				} 
				break;
			}
			case 'P' : {
				if (i+1<argc && argv[i+1][0]!='-')
				{
					// read name of postscript file
					ps_file_name = std::string(argv[i+1]);
					ps_output = true;
					i++;
					num_of_opts++;
				} 
				break;
			}
			case 't' : {
				if (strlen(argv[i])>2)
					if (argv[i][2]=='c')
					{
						count_results = true;
					} 
				if (i+1<argc)
				{
					threshold = atof(argv[i+1]);
					i++;
					num_of_opts++;
				} else {
					params_good = false;		
					threshold = 0.;
				}
				break;
			}
			case 'm' : {
				use_multisets = true;
				break;
			}
			case 'w' : {
				weighted = true;
				weight_type = 1;
				if (strlen(argv[i])>2)
					switch (argv[i][2])
					{
						case 'm' :
							weight_type = 1;
							break;
						case 'p':
							weight_type = 2;
							break;
						default:
							std::cout<<"ignoring unknown weight type "<<argv[i][2]<<std::endl;
							weight_type = 1;
							break;
					}
				
				break;
			}
			default: {
				std::cout<<"ignoring unknown option "<<argv[i][1]<<std::endl;
			}
		}
	}
	return params_good;
}

void print_help_message()
{
	std::cout<<"Use bbq as follows:\n\n";
	std::cout<<"bbq [options] <L> <footprint-file> <genome1> ... <genomeK>,\n\n";
	std::cout<<"where <L> is an integer indicating the cluster length to optimize\n";
	std::cout<<"for, <footprint-file> is a filename as well as <genome1> ... <genomeK>.\n";
	std::cout<<"valid options are:\n";
	std::cout<<"-g <S> for grouping, where <S> is an integer indicating the length of the group prefix for fragment names;\n";
	std::cout<<"-h <N> computing the best N matches;\n";
	std::cout<<"-S (Stupid user) if the output is supposed to be more verbose so that it can be understood by a stupid user :)\n";
	std::cout<<"-m use multisets (i.e., multiple occurences in 1 cluster make a difference)\n";
	std::cout<<"-u (unweighted) to switch off weighting and optimize for intersection cardinality;\n";
	std::cout<<"-w<t> select weighting based on mismatches (<t>=m) or based on p-values (<t>=p)\n";
	std::cout<<"-t <T> for thresholding, with T denoting a float threshold. All occurences whose weight (or intersection cardinality, in case -u was specified) exceeds T are printed.\n";
	std::cout<<"-P <filename> for postscript output to file <filename>\n";
	std::cout<<"-p display progress information\n";
	std::cout<<"-r include reverse complemented fragments so that both 3' and 5' occurences may be part of a cluster\n";
	std::cout<<"-A<x> with <x>=1 or <x>=2 for choosing Algorithm (A1) (default) or (A2), respectively.\n";
	std::cout<<"-D <delta> for using a bounded-difference algorithm, where <delta> is an integer specifiying the difference bound.\n";
	std::cout<<"\n";
}

void read_data(int argc, char** argv)
{
	
	int j,l;
	params_good = true;

	if (argc-num_of_opts<4)
	{
		print_help_message();
		params_good = false;
	}
	
	// 0. READ CLUSTER LENGTH L
	L = atoi(argv[num_of_opts+1]);
	
	// I. READ FRAGMENT SEQUENCES s_1,...,s_k from the fragment-file
	std::list<char*> s_list, s_name_list;
	std::list<int> n_list, m_list;
	std::ifstream frag_file(argv[num_of_opts+2]);
	skip(frag_file);	
	while (frag_file.good())
	{
		char read_buf[1024];
		// read fragment name
		read_buf[0]=0;
		read_buf[1023] = 0;
		int name_len;
		frag_file.getline(read_buf,1024,' ');
		name_len = strlen(read_buf);
		char* curr_name = new char[name_len+1];
		strcpy(curr_name,read_buf);
		s_name_list.push_back(curr_name);
		
		skip(frag_file);
		
		// read fragment sequence, currently delimited to 1024 characters
		read_buf[0]=0;
		read_buf[1023] = 0;
		int s_len;
		frag_file.getline(read_buf,1024,' ');
		s_len = strlen(read_buf);
		char* curr_s = new char[s_len+1];
		strcpy(curr_s,read_buf);
		s_list.push_back(curr_s);
		n_list.push_back(s_len);
		skip(frag_file);
		
		// read maximum edit distance for the fragment
		read_buf[0]=0;
		read_buf[1023] = 0;
		int d_len;
		frag_file.getline(read_buf,1024);
		m_list.push_back(atoi(read_buf));
		skip(frag_file);		
	}

	k = s_list.size();
	if (reverse)
		k *= 2;
	s = new char*[k];
	m = new int[k];
	n = new int[k];
	s_names = new char*[k];
	
	std::list<char*>::iterator s_it = s_list.begin();
	std::list<char*>::iterator names_it = s_name_list.begin();
	std::list<int>::iterator n_it = n_list.begin();
	std::list<int>::iterator m_it = m_list.begin();
	for (j=0; j<k; j++, ++s_it, ++n_it, ++names_it, ++m_it)
	{
		n[j] = *n_it;
		m[j] = *m_it;
		s[j] = *s_it;
		s_names[j] = *names_it;
		int iota;
		char* s_j = s[j];
		int n_j = n[j];
		for (iota=0; iota<n_j; iota++)
			s_j[iota] = toupper(s_j[iota]);
	}
	
	// II. READ GENOME SEQUENCES T_1,...,T_K
	K = argc-3-num_of_opts;
	T = new char*[K];
	N = new int[K];
	for (int i=num_of_opts+3; i<argc; i++)
	{
		read_genome(argv[i],i-3-num_of_opts);
	}

	if (reverse)
		fill_reverse_segments();
	
}

void fill_reverse_segments()
{

	if (!reverse)
		return;
	
	int i;
	
	char* compl_table = new char[256];

	for(i=0; i<256; i++) compl_table[i]=(int)'-';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';
	compl_table[(int)'R']='Y';
	compl_table[(int)'Y']='R';
	compl_table[(int)'M']='K';
	compl_table[(int)'K']='M';
	compl_table[(int)'S']='S';
	compl_table[(int)'W']='W';
	compl_table[(int)'H']='D';
	compl_table[(int)'B']='V';
	compl_table[(int)'V']='B';
	compl_table[(int)'D']='H';
	compl_table[(int)'N']='N';
	compl_table[(int)'a']='t';
	compl_table[(int)'c']='g';
	compl_table[(int)'g']='c';
	compl_table[(int)'t']='a';
	compl_table[(int)'r']='y';
	compl_table[(int)'y']='r';
	compl_table[(int)'m']='k';
	compl_table[(int)'k']='m';
	compl_table[(int)'s']='s';
	compl_table[(int)'w']='w';
	compl_table[(int)'h']='d';
	compl_table[(int)'b']='v';
	compl_table[(int)'v']='b';
	compl_table[(int)'d']='h';
	compl_table[(int)'n']='n';
	
	int j, iota;
	// k is divisible by two, since it has been mult by two due to
	// reverse option before.
	int k_2 = k/2;
	for (j=0; j<k_2; j++)
	{
		int n_j = n[j];
		s[k_2+j] = new char[n[j]+1];
		n[k_2+j] = n[j];
		s_names[k_2+j] = new char[strlen(s_names[j])+5];
		strcpy(s_names[k_2+j],s_names[j]);
		strcat(s_names[k_2+j],"-rev");
		for (iota=0; iota<n_j; iota++)
			s[k_2+j][iota] = compl_table[s[j][n_j-iota-1]];
		s[k_2+j][n_j] = (char)0;
	}
	//for (j=0; j<k; j++)
	//	std::cout<<s_names[j]<<" "<<s[j]<<std::endl;
}


void free_all()
{
	int i,j;
	for (i=0; i<K; i++)
		delete[] T[i];
	delete[] T;
	T = NULL;;
	for (j=0; j<k; j++)
		delete[] s[j];
	delete[] s;
	s = NULL;
	for (j=0; j<k; j++)
		delete[] s_names[j];
	delete[] s_names;
	s_names = NULL;
	delete[] n;
	n = NULL;
	delete[] N;
	N = NULL;
	delete[] m;
	m = NULL;
	
}

void test_unweighted()
{
	char** T = new char*[2];
	T[0] = new char[17];
	T[1] = new char[20];	
	memcpy(T[0],"dbccabcdddabcdba",17);
	memcpy(T[1],"abcaadabcbccddbabcc",20);
	char** s = new char*[3];
	s[0] = new char[4];
	s[1] = new char[4];	
	s[2] = new char[4];	
	memcpy(s[0],"abc",4);
	memcpy(s[1],"bcc",4);
	memcpy(s[2],"dab",4);

	int n[3] = {3,3,3};
	int N[3] = {16,19};
	footprint_detector fd(3,2,s,T,n,N);
	fd.run(6);
	std::cout<<"The maximum L-occurence found has size "<<fd.get_result()<<"."<<std::endl;
	fd.print_result();
	delete[] s[0];
	delete[] s[1];
	delete[] s[2];
	delete[] s;
	delete[] T[0];
	delete[] T[1];
	delete[] T;
}


/// Use bbq as follows:
///
/// bbq <L> <footprint-file> <genome1> ... <genomeK>,
///
/// where <L> is an integer indicating the cluster length to optimize
/// for, <footprint-file> is a filename as well as <genome1> ... <genomeK>.
int main(int argc, char** argv)
{
	
	T = NULL;;
	s = NULL;
	s_names = NULL;
	n = NULL;
	N = NULL;
	m = NULL;
	std::cout<<"\nWelcome to the Best Barbeque tool.\n\n";

	read_parameters(argc,argv);
	if (!params_good)
	{
		print_help_message();
		return 0;
	}
	read_data(argc,argv);
	if (!params_good)
	{
		print_help_message();
		return 0;
	}
	
	if (threshold>=0. && ps_output && !count_results)
		std::cerr<<"warning: combining options -t and -P may produce large number of output files.\n\n";
	
	if (!use_multisets)
	{
		w_set_footprint_detector wfd(k,K,s,s_names,T,&(argv[num_of_opts+3]),n,N,m);
		wfd.set_weighted(weighted);
		wfd.set_count_results(count_results);
		wfd.set_weight_type(weight_type);
		wfd.set_reverse(reverse);
		wfd.set_progress_report(report_progress);
		wfd.set_algorithm(algorithm);
		wfd.set_number_of_hits(num_of_hits);
		wfd.set_grouping(grouped);
		wfd.set_output_detail(stupid_user);
		wfd.set_group_characters(num_of_group_characters);
		wfd.set_postscript(ps_output);
		wfd.set_ps_filename(ps_file_name);
		if (bounded_difference)
			wfd.set_delta(delta);
		wfd.run(L,threshold);
		if (threshold<0.)
			wfd.print_result();
		wfd.clear();
		free_all();
	} else {
		if (!weighted)
		{
			w_mset_footprint_detector wfd(k,K,s,s_names,T,&(argv[num_of_opts+3]),n,N,m);
			wfd.set_weighted(weighted);
			wfd.set_count_results(count_results);
			wfd.set_weight_type(weight_type);
			wfd.set_algorithm(algorithm);
			wfd.set_number_of_hits(num_of_hits);
			wfd.set_grouping(grouped);
			wfd.set_output_detail(stupid_user);
			wfd.set_group_characters(num_of_group_characters);
			wfd.set_postscript(ps_output);
			wfd.set_ps_filename(ps_file_name);
			if (bounded_difference)
				wfd.set_delta(delta);
			wfd.run(L,threshold);
			if (threshold<0.)
				wfd.print_result();
			wfd.clear();
			free_all();
		} else {
			w_wmset_footprint_detector wfd(k,K,s,s_names,T,&(argv[num_of_opts+3]),n,N,m);
			wfd.set_weighted(weighted);
			wfd.set_count_results(count_results);
			wfd.set_weight_type(weight_type);
			wfd.set_algorithm(algorithm);
			wfd.set_number_of_hits(num_of_hits);
			wfd.set_grouping(grouped);
			wfd.set_output_detail(stupid_user);
			wfd.set_group_characters(num_of_group_characters);
			wfd.set_postscript(ps_output);
			wfd.set_ps_filename(ps_file_name);
			if (bounded_difference)
				wfd.set_delta(delta);
			wfd.run(L,threshold);
			if (threshold<0.)
				wfd.print_result();
			wfd.clear();
			free_all();
		}
	}
	
	return 0;
	
}

//} // end namespace bbq

