// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#ifndef WEIGHTED_BITMULTISETS_H
#define WEIGHTED_BITMULTISETS_H

#include <iostream>


namespace bbq {

/// This class provides the most advanced (but also most inefficient)
/// way of handling sets, which is a major part of the algorithms
/// underlying \c bbq. This class particularly supports weighted
/// sets whose cardinality is unlimited. It supports weighing as well
/// as multiset capabilities, but is far less memory efficient than
/// the unweighted and simple-set versions.
class weighted_bit_multi_set 
{
public:

	typedef unsigned char c_uint;

	c_uint* counts;
	
	/// k corresponds to the number of bits and hence the number of
	/// sequence fragments in footprint_detector
	static int k;

	/// left for compatibility reasons with other set types
	static bool weighted;	
	static int k_;
	float* weights;

	/// copy constructor
	weighted_bit_multi_set(const weighted_bit_multi_set&);

	/// assignment operator
	weighted_bit_multi_set& operator=(const weighted_bit_multi_set& vb);

	/// default constructor
	weighted_bit_multi_set();

	/// destrcutor
	~weighted_bit_multi_set()
		{
			delete[] counts;
		}

	bool subseteq(const weighted_bit_multi_set&) const;
	
	// this version of subseteq checks only up to the J-th bit whether
	// *this is a subset of the set passed as an argument.
	bool subseteq(const weighted_bit_multi_set&, const unsigned int& J) const;
	
	/// We obtain a canonical order on subsets of [1:k] by considering
	/// such a set as a number between 0 and 2^k-1. This order is
	/// needed for inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator<(const weighted_bit_multi_set&) const;

	/// Two sets are considered equivalent iff they contain the same
	/// elements, not regarding their weights. this is needed for
	/// inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator==(const weighted_bit_multi_set&) const;

	/// returns the number of elements contained in the multiset
	int cardinality();

	/// in the simple form implemented here, the weight is equal to the
	/// number of elements contained in the multiset.
	float weight();

	friend std::ostream& operator<<(std::ostream& o,const weighted_bit_multi_set&);

	/// standard multiset intersection
	void intersect(const weighted_bit_multi_set& b);

	/// add one element j to the multiset. The weight is (currently)
	/// neglected.
	void set(const int& j, const float& w);

	/// change the weight of element j; only needed for sets, not for multisets.
	void reset(const int& j, const float& w);

	/// remove ONE element j from the multiset.
	void decrease(const int& j, const float& w);

	/// remove ALL elements j from the multiset. Can actually be
	/// ignored for multisets, since the remove job is taken by
	/// decrease.
	void remove(const int& j, const float& w);

	/// is there at least one element with the index passed by the
	/// parameter?
	bool contains(const int&);
	
};

} // end namespace bbq


#endif

