#ifndef WEIGHTED_BITSETS_H
#define WEIGHTED_BITSETS_H

#include <iostream>

namespace bbq {


/// This class provides one possible way of handling sets, which is a
/// major part of the algorithms underlying \c bbq. This class
/// particularly supports weighted sets of arbitrary cardinality, but
/// no multiset capabilities.
class weighted_variable_bitset 
{
public:

	unsigned int* bits;
	float* weights;
	
	/// k corresponds to the number of bits and hence the number of
	/// sequence fragments in footprint_detector
	static int k;

	/// k corresponds to the number of words needed to store all bits,
	/// i.e., k_ = k/32+1 (or (k>>5)+1) on a 32-bit machine.
	static int k_;

	/// turn weighting features on and off. Not to be changed during a
	/// \c run() of weighted_footprint_detector.
	static bool weighted;
	
	/// copy constructor
	weighted_variable_bitset(const weighted_variable_bitset&);

	/// assignment operator
	weighted_variable_bitset& operator=(const weighted_variable_bitset& vb);
	
	/// default constructor
	weighted_variable_bitset();

	/// destrcutor
	~weighted_variable_bitset();
	
	/// Test whether \c this is a subset of another set.
	bool subseteq(const weighted_variable_bitset& b) const;

	// this version of subseteq checks only up to the J-th bit whether
	// *this is a subset of the set passed as an argument.
	// This gives a slight speedup for large values of k.
	bool subseteq(const weighted_variable_bitset&, const unsigned int& J) const;
	
	/// returns the number of elements contained in the set
	int cardinality();

	/// sum of all weights associated with the elements contained in
	/// the set
	float weight();

	friend std::ostream& operator<<(std::ostream& o,const weighted_variable_bitset&);
	
	/// standard set intersection with adding weights associated with
	/// the elements that are contained in the resulting intersection
	void intersect(const weighted_variable_bitset& b);

	/// add one element \c j to the set with associated weight \c w.
	void set(const int& j, const float& w);

	/// change the weight of element j; only needed for sets, not for multisets.
	void reset(const int& j, const float& w);

	/// only needed for compatibility with the multiset supporting classes
	void decrease(const int& j, const float& w);

	/// remove index j from the set.
	void remove(const int& j, const float& w);

	/// return \c true iff the element carrying the given index is
	/// present in the set.
	bool contains(const int&);

	/// We obtain a canonical order on subsets of [1:k] by considering
	/// such a set as a number between 0 and 2^k-1. This order is
	/// needed for inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator<(const weighted_variable_bitset&) const;

	/// Two sets are considered equivalent iff they contain the same
	/// elements, not regarding their weights. this is needed for
	/// inserting weighted sets into (multi-)sets for the \c
	/// enumerate_differences() method. Note that weightings are not
	/// taken into account; two sets containing the same elements but
	/// different weights are considered equivalent (which is the
	/// reason why they are stored in a multiset rather than a set).
	bool operator==(const weighted_variable_bitset&) const;	
	
};


} // end namespace bbq


#endif

