// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#include <limits.h>
#include <float.h>
#include <stdlib.h>

#include <iostream>
#include <fstream>
#include <ios>
#include <math.h>

#include <list>
#include <vector>
#include <string>
#include <algorithm>
#include "triple.h"
#include "bbq_postscript.h"

#include "weighted_footprint_detector.h"


namespace bbq {
	
/// A cluster triple is a triplet (p,e,j), where p is the position, e
/// the edit distance and j the fragment index of an occurence in some
/// T_i.
typedef std::triple< int,int,int > cluster_triple;
typedef std::set<cluster_triple> cluster_set;


#define max(x,y) ( ( (x) > (y) ) ? (x) : (y) )

typedef std::list< std::pair<int,float> > open_interval_list;

// ==============================================================
template<class w_bit_set>
weighted_footprint_detector<w_bit_set>::weighted_footprint_detector()
// ==============================================================
{
	enum_differences = false;	
	reverted = false;
	algorithm = 1;
	
	// numbers of fragments, groups and genomes
	k = 0;
	kappa = 0;
	K = 0;

	// memory for cell enumeration and backtracking
	s = NULL;
	s_names = NULL;
	T = NULL;
	T_names = NULL;
	n = NULL;
	N = NULL;
	m = NULL;

	// optimal solution memory
	best_nu = NULL;
	last_nu = NULL;

	// number of hits
	h = 1;

	// memory neccessary for grouping stuff
	group_names = NULL;
	group_sizes = NULL;
	group_indices = NULL;
	
}


// ==============================================================
template<class w_bit_set>
weighted_footprint_detector<w_bit_set>::weighted_footprint_detector(int k_, int K_, char** s_, char** s_names_, char** T_, char** T_names_, int* n_, int* N_, int* m_)
// ==============================================================
{
	enum_differences = false;	
	reverted  = false;
	algorithm = 1;
	k = k_;
	K = K_;
	s = s_;
	s_names = s_names_;
	T = T_;
	T_names = T_names_;
	n = n_;
	N = N_;
	m = m_;	
	best_nu = NULL;
	last_nu = NULL;

	w_bit_set::k = k;
	w_bit_set::k_ = (k>>5)+1;
	// re-init best_intersection with the new static variables k and
	// k_.
	best_intersection = w_bit_set();
	
	// number of hits
	h = 1;

	// memory neccessary for grouping stuff
	group_names = NULL;
	group_sizes = NULL;
	group_indices = NULL;

	compute_index_table();
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::init_arrangement_cells()
// ==============================================================
{
	
	int i, j, ii, jj;
	
	int* open_interval_counter = new int[k];
	int* nu = new int[k];	
	for (i=0; i<K; i++)
	{
		
		p[i] = new int*[k];
		M[i] = new int*[k];
		R[i] = new bool*[k];
		lambda[i] = new int[k];
		Lambda[i] = 0;

		w_bit_set current_cell_set;

		// count how many intervals of color j are open
		int lambda_i = 0;
		
		for (j=0; j<k; j++)
		{
			nu[j] = 0;
			open_interval_counter[j] = 0;
			// compute all lambda[i][j] many occurences of s[j] in T[i].
			//if (grouped)
			get_group_occurences(i,j);
			//else
			//get_occurences(i,j);
			lambda_i += lambda[i][j];
		}
		
		// lambda_i contains the total number of occurences of fragments
		// in T_i. Apparently, 2*lambda_i is an upper bound for the
		// number of cells (non-empty) we may encounter.
		C[i] = new w_bit_set[2*lambda_i];
		P[i] = new int[2*lambda_i];

		// yes, we're SWEEPING again!
		int sweep_position = 0;
		int occs_done = 0; // how many p[i][j]s have we sweeped over completely so far?

		// open_intervals[j] contains the list of open intervals of
		// color j. These interval ends are ordered accoring to their
		// positions.
		std::vector< open_interval_list > open_intervals(k);
		int total_occs = 0;
		for (j=0; j<k; j++)
		{
			open_intervals[j].clear();
			total_occs += lambda[i][j];
		}
		
		while (occs_done<total_occs)
		{
			// determine the starting position of the next cell.

			// determine the next BEGINNING of an interval (using p[i][j][nu[j]])
			int best_beg_j = 0;
			int best_beg_p = INT_MAX;
			float best_beg_w = 0.;
			float best_end_w = 0.;
			for (j=0; j<k; j++)
			{
				if (nu[j]<lambda[i][j])
				{
					if (p[i][j][nu[j]] < best_beg_p)
					{
						best_beg_p = p[i][j][nu[j]];
						best_beg_w = get_weight(i,j,nu[j]);//((float)(m[j]-M[i][j][nu[j]])+1.) / (((float)(m[j])+1.));
						best_beg_j = j;
					}
				}				
			}

			// determine the next END of an interval
			// (which can be done using the first entries of the k open_intervals lists)
			int best_end_j = 0;
			int best_end_p = INT_MAX;
			int last_k;
			for (j=0; j<k; j++)
			{
				if (open_intervals[j].size()>0)
				{
					if ((open_intervals[j].begin()->first) < best_end_p)
					{
						best_end_p = open_intervals[j].begin()->first;
						best_end_w = open_intervals[j].begin()->second;
						best_end_j = j;
					}
				}				
			}

			// Which occurs earlier: the next BEGINNING interval or the
			// next ENDING interval?
			if (best_beg_p <= best_end_p)
			{
				//std::cout<<"b"<<best_beg_j<<"\n";
				sweep_position = best_beg_p;
				nu[best_beg_j]++;
				open_interval_counter[best_beg_j]++;
				// activate the inteval's color in the set of currently active colors
				if (best_beg_w>current_cell_set.weights[best_beg_j])
					current_cell_set.set(best_beg_j,best_beg_w);
				else
					current_cell_set.set(best_beg_j,current_cell_set.weights[best_beg_j]);
				// mark the interval as being opened by appending it to
				// the intervals_open[j] list.
				open_intervals[best_beg_j].push_back(std::pair<int,float>(best_beg_p + L,best_beg_w));
			} else {
				//std::cout<<"e"<<best_end_j<<"\n";
				sweep_position = best_end_p;
				open_interval_counter[best_end_j]--;
				// unmark the interval by removing it from the
				// intervals_open[j] list.
				open_intervals[best_end_j].pop_front();
				// for handling multisets, we need a decrease method for
				// feature sets as opposed to the remove method, which is
				// there for standard sets.
				current_cell_set.decrease(best_end_j,best_end_w);				
				if (open_interval_counter[best_end_j]<=0)
					// deactivate the inteval's color in the set of currently active colors
					current_cell_set.remove(best_end_j,best_end_w);
				else {
					// there is at least one interval of color best_end_j
					// acitve.  OUCH... we have to update the weight for
					// color best_end_j. To this end, we search the whole
					// list open_intervals[best_end_j] for the best
					// remaining weight.
					if (open_intervals[best_end_j].size()>0)
					{
						open_interval_list::iterator oi_it = open_intervals[best_end_j].begin();
						open_interval_list::iterator oi_end = open_intervals[best_end_j].end();
						float new_best_w = 0.;
						for (; oi_it!=oi_end; oi_it++)
							if (oi_it->second > new_best_w)
								new_best_w = oi_it->second;
						current_cell_set.reset(best_end_j,new_best_w);
					}
				}
				occs_done++;
			}
			// What The F*#!ing Hell is gcc doing with the assignment operator call?
			C[i][Lambda[i]] = w_bit_set(current_cell_set);
			P[i][Lambda[i]] = sweep_position;
			Lambda[i]++;
		}
		//std::cout<<"*** "<<i<<"\n";
		/*for (int jj=0; jj<Lambda[i]; jj++)
		std::cout<<P[i][jj]<<" "<<C[i][jj]<<std::endl;*/
	}
	///std::cout<<"*** "<<i<<"\n";
	delete[] open_interval_counter;
	delete[] nu;
}


// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::w_bound(const w_bit_set& M, const unsigned int& j)
// ==============================================================
{

	int i,nu;
	bool found;
	float last_best_w = 0.;

	for (i=0; i<K; i++)
	{
		float best_w = 0.;
		// find nu sucht that C[i][nu] contains M and has maximum weight.
		found = false;
		w_bit_set* C_i = C[i];
		// in order to get fast access to C[i][nu], we do some fancy
		// little pointer arithmetic...

		// yes, this procedure is one of the most time critical ones, so
		// it should be worth the effort...
		w_bit_set* C_i_nu = &(C[i][0]);
		w_bit_set* C_i_Lambda_i = &(C[i][Lambda[i]]);
		for (nu=0; C_i_nu != C_i_Lambda_i; C_i_nu++, nu++)
			if (M.subseteq(*C_i_nu,j))
			{
				found = true;
				if (j==k)
				{
					// opposed to the unweighted case, we cant break at
					// this point, since we might find a better weight in
					// C_i. If this happens, best_w needs an
					// update. Fortunately, it suffices to do so if j==k,
					// i.e., if we need to compute a complete solution
					// candidate for enum_w_sets().
					w_bit_set intersection = M;
					intersection.intersect(*C_i_nu);
					float curr_w = intersection.weight();
					if (curr_w>best_w)
					{
						last_nu[i] = nu;
						best_w = curr_w;
					}
				} else {
					last_nu[i] = nu;
					break;
				}
			}
		last_best_w += best_w;
		if (!found)
			// if no suitable subsets of M were found, then the
			// bound for the branch-and-bound-like algorithm indeed
			// holds, so that we return true.
			return true;
	}
	anything_found = true;
	if (j==k)
		if (last_best_w>=overall_best_weight)
		{
			// if suitable subsets of M were found and, in case j==k, their
			// weight aditionally is better than the best solution produced
			// so far, then the bound for the branch-and-bound like
			// algorithm does not hold, so that we return false.
			
			best_intersection = M;
			best_nu[0] = last_nu[0];
			for (i=1; i<K; i++)
			{
				best_nu[i] = last_nu[i];
				best_intersection.intersect(C[i][last_nu[i]]);
			}	
			overall_best_weight = last_best_w;
			if (threshold>=0.)
			{
				if (!count_results)
					print_result();
				num_of_solutions++;
				overall_best_weight = threshold;
			}

			// if threshold>=0., overall_bes_weight is always equal to
			// threshold and will never changed!
			return false;
		} else {
			return true;
		}
	return false;
	
}

// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::h_best_w_bound(const w_bit_set& M, const unsigned int& j)
// ==============================================================
{

	int i,nu;
	bool found;
	float last_best_w = 0.;

	for (i=0; i<K; i++)
	{
		float best_w = 0.;
		// find nu sucht that C[i][nu] contains M and has maximum weight.
		found = false;
		w_bit_set* C_i = C[i];
		// in order to get fast access to C[i][nu], we do some fancy
		// little pointer arithmetic...

		// yes, this procedure is one of the most time critical ones, so
		// it should be worth the effort...
		w_bit_set* C_i_nu = &(C[i][0]);
		w_bit_set* C_i_Lambda_i = &(C[i][Lambda[i]]);
		for (nu=0; C_i_nu != C_i_Lambda_i; C_i_nu++, nu++)
			if (M.subseteq(*C_i_nu,j))
			{
				found = true;
				if (j==k)
				{
					// opposed to the unweighted case, we cant break at
					// this point, since we might find a better weight in
					// C_i. If this happens, best_w needs an
					// update. Fortunately, it suffices to do so if j==k,
					// i.e., if we need to compute a complete solution
					// candidate for enum_w_sets().
					w_bit_set intersection = M;
					intersection.intersect(*C_i_nu);
					float curr_w = intersection.weight();
					if (curr_w>best_w)
					{
						last_nu[i] = nu;
						best_w = curr_w;
					}
				} else {
					last_nu[i] = nu;
					break;
				}
			}
		last_best_w += best_w;
		if (!found)
			// if no suitable subsets of M were found, then the
			// bound for the branch-and-bound-like algorithm indeed
			// holds, so that we return true.
			return true;
	}
	anything_found = true;
	if (j==k)
	{
		w_bit_set tmp_intersection = M;
		best_nu[0] = last_nu[0];
		for (i=1; i<K; i++)
		{
			tmp_intersection.intersect(C[i][last_nu[i]]);
		}	
		update_h_best_list(last_best_w,last_nu,tmp_intersection);
	}
	return false;
	
}

// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::bound(const w_bit_set& M, const unsigned int& j)
// ==============================================================
{

	int i,nu;
	bool found;

	//all_found = true;
	for (i=0; i<K; i++)
	{
		// check whether there is a j such that C[i][j] contains M.
		found = false;
		// for getting the best possible speed, do some iterator-like
		// pointer arithmetic...
		w_bit_set* C_i = C[i];
		w_bit_set* C_i_nu = &(C_i[0]);
		w_bit_set* C_i_end = &(C_i[Lambda[i]]);
		for (nu=0; C_i_nu != C_i_end; C_i_nu++, nu++)
			if (M.subseteq(*C_i_nu,j))
			{
				last_nu[i] = nu;
				found = true;
				break;
			}
		if (!found)
			// if no suitable subsets of M were found, then the
			// bound for the branch-and-bound-like algorithm indeed
			// holds, so that we return true.
			return true;
	}
	
	anything_found = true;
	// if suitable subsets of M were found, then the bound for the
	// branch-and-bound like algorithm does not hold, so that we return
	// false.
	return false;
	
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enum_w_sets(w_bit_set M, int j)
// ==============================================================
{
	int i;
	
	if (j==k)
	{
		// w_bound does all update steps for us, in case an improved
		// solution is found.
		w_bound(M,j);
		return;
	}
	if (!w_bound(M,j))
	{
		M.set(j,0.);
		enum_w_sets(M,j+1);
		M.remove(j,0.);
		enum_w_sets(M,j+1);
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enum_h_best_w_sets(w_bit_set M, int j)
// ==============================================================
{
	int i;
	
	if (j==k)
	{
		// w_bound does all update steps for us, in case an improved
		// solution is found.
		h_best_w_bound(M,j);
		return;
	}
	if (!h_best_w_bound(M,j))
	{
		M.set(j,0.);
		enum_h_best_w_sets(M,j+1);
		M.remove(j,0.);
		enum_h_best_w_sets(M,j+1);
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enum_sets(w_bit_set M, int j)
// ==============================================================
{
	int i;

	//std::cout<<"enum_sets "<<M<<"\n";
	if (j==k)
	{
		if (!bound(M,j))
		{
			float wght = M.weight();
			if (wght>overall_best_weight)
			{
				best_intersection = M;
				for (i=0; i<K; i++)
				{
					best_nu[i] = last_nu[i];
					//best_intersection.intersect(C[i][last_nu[i]]);
				}
				overall_best_weight = wght;
				if (threshold>=0)
				{
					if (!count_results)
					print_result();
					overall_best_weight = threshold;
					num_of_solutions++;
				}
			}
		}
		if (report_progress)
			progress++;
	} else if (!bound(M,j))	{
		M.set(j,1.);
		enum_sets(M,j+1);
		M.remove(j,1.);
		enum_sets(M,j+1);
	} else if (report_progress) {
		// update progress information:
		// 2^(k-j) instances are "bounded away".
		mpz_class d, f;
		
		d = 1;
		f = d<<(k-j);
		progress += f;		
	}
	if (report_progress)
		if (prgr_count==1024)
		{
			mpf_class x(progress);
			mpf_class y(num_of_inst);
			mpf_class z(num_of_inst);
			z = x/y;
			z *= mpf_class(100.);
			std::cerr<<(char)13;
			std::cerr<<"progress: "<<z<<"%                        ";
			std::cerr.flush();
			prgr_count = 0;
		} else {
			prgr_count++;
		}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enum_h_best_sets(w_bit_set M, int j)
// ==============================================================
{
	int i;

	//std::cout<<"enum_sets "<<M<<"\n";
	if (j==k)
	{
		if (!bound(M,j))
		{
			float wght = M.weight();
			update_h_best_list(wght,last_nu,M);
		}
		if (report_progress)
			progress++;
	} else if (!bound(M,j))	{
		M.set(j,1.);
		enum_h_best_sets(M,j+1);
		M.remove(j,1.);
		enum_h_best_sets(M,j+1);
	} else if (report_progress) {
		// update progress information:
		// 2^(k-j) instances are "bounded away".
		mpz_class d, f;
		
		d = 1;
		f = d<<(k-j);
		progress += f;		
	}
	if (report_progress)
		if (prgr_count==1024)
		{
			mpf_class x(progress);
			mpf_class y(num_of_inst);
			mpf_class z(num_of_inst);
			z = x/y;
			z *= mpf_class(100.);
			std::cerr<<(char)13;
			std::cerr<<"progress: "<<z<<"%                        ";
			std::cerr.flush();
			prgr_count = 0;
		} else {
			prgr_count++;
		}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enumerate_subsets()
// ==============================================================
{
	if (report_progress)
	{
		std::cerr<<(char)13;
		std::cerr<<"progress: 0%                        ";
		std::cerr.flush();
	}
	anything_found = false;
	if (threshold<0.)
		overall_best_weight = 0.;
	else
		overall_best_weight = threshold;
	//std::cout.flags ( ios_base::fixed );
	std::cerr.precision(32);
	if (report_progress)
	{
		progress = 0;
		prgr_count = 0;
		num_of_inst = 1;
		num_of_inst = num_of_inst<<k;
	}
	// default constructor of w_bit_set is expected to yield the empty
	// set!
	w_bit_set M;
	if (!weighted)
		enum_sets(M,0);
	else
		enum_w_sets(M,0);
	if (report_progress)
	{
		std::cerr<<(char)13;
		std::cerr.flush();
		std::cerr<<"progress: 100%                                           \n";
		std::cerr.flush();
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enumerate_h_best_subsets()
// ==============================================================
{

	// we dont deal with thresholding, so we can set the h-th best weight to 0.
	h_best_weight = 0.;
	
	// the number of solutions found so far is nadazero.
	h_so_far = 0;
	
	if (report_progress)
	{
		std::cerr<<(char)13;
		std::cerr<<"progress: 0%                        ";
		std::cerr.flush();
		anything_found = false;
	}
	if (threshold<0.)
		overall_best_weight = 0.;
	else
		overall_best_weight = threshold;
	//std::cout.flags ( ios_base::fixed );
	std::cerr.precision(32);
	if (report_progress)
	{
		progress = 0;
		prgr_count = 0;
		num_of_inst = 1;
		num_of_inst = num_of_inst<<k;
	}
	// default constructor of w_bit_set is expected to yield the empty
	// set!
	w_bit_set M;
	if (!weighted)
		enum_h_best_sets(M,0);
	else
		enum_h_best_w_sets(M,0);
	if (report_progress)
	{
		std::cerr<<(char)13;
		std::cerr.flush();
		std::cerr<<"progress: 100%                                           \n";
		std::cerr.flush();
	}
}


// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::increase(int* nu, int I)
// ==============================================================
{
	int i;
	bool nu_0_full = (nu[0]==Lambda[0]-1);
	for (i=K-1; i>=0; i--)
	{
		++nu[i];
		if (nu[i]==Lambda[i])
			nu[i] = 0;
		else 
			break;
	}
	return (nu[0]==0 && nu_0_full);
}
	

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::compute_intersections()
// ==============================================================
{
	int i, mu;
	int C_size = 1;
	int* nu = new int[K];

	if (report_progress)
	{
		std::cerr.precision(32);		
		progress = 0;
		prgr_count = 0;
		num_of_inst = 1;
		for (i=0; i<K; i++)
			num_of_inst *= (double)(Lambda[i]);
		std::cerr<<(char)13;
		std::cerr<<"progress: 0%                        ";
		std::cerr.flush();
	}
	num_of_inst--;
	
	if (threshold<0.)
		overall_best_weight = 0.;
	else
		overall_best_weight = threshold;

	for (i=0; i<K; i++)
	{
		nu[i] = 0;
		if (Lambda[i]==0)
		{
			anything_found = false;
			return;
		}
	}

	anything_found = true;	
	bool done = false;
	while (!done)
	{
		if (report_progress)
			if (prgr_count==65536)
			{
				mpf_class x(num_of_inst);
				mpf_class y(num_of_inst);
				mpf_class z(num_of_inst);
				x = nu[0];
				for (i=1; i<K; i++)
				{
					x *= (Lambda[i]);
					x += (nu[i]);
				}
				z = x/y;
				z *= mpf_class(100.);
				std::cerr<<(char)13;
				std::cerr<<"progress: "<<z<<"%                        ";
				std::cerr.flush();
				prgr_count = 0;
			} else {
				prgr_count++;
			}
		// compute the intersection...
		w_bit_set intersection = C[0][nu[0]];
		bool bound = false;
		float wght;
		for (i=1; i<K; i++)
		{
			intersection.intersect(C[i][nu[i]]);
			// The following bound step is based on the premise that 1.0
			// is an upper bound for the weight of a match, i.e., an
			// exact occurence of s_j in T_i is weighted by 1.0 and all
			// approximate occurence are weighted lower than 1.0.
			wght = intersection.weight();
			if (wght+(float)(K-i+1)<overall_best_weight)
			{
				// perform a bound-step in the branch-and-bound-algorithm
				done = increase(nu,i);
				bound = true;
				break;
			}
		}

		if (!bound)
		{
			if (wght>=overall_best_weight)
			{
				// if we reach this point, we have
				//
				//    intersection.cardinality() > threshold,
				//
				// where intersection is the intersection of all C[i][nu[i]],
				// for all i in [1:K]. In other words, this means that we
				// have found a new valid intersection.			
				w_bit_set saved_intersection = intersection;

				best_intersection = saved_intersection;
				for (int ii=0; ii<K; ii++)
					best_nu[ii] = nu[ii];
				if (threshold<0.)
					overall_best_weight = intersection.weight();
				else
				{
					num_of_solutions++;
					overall_best_weight = intersection.weight();
					if (!count_results)
						print_result();
					overall_best_weight = threshold;
				}
			}
			// finally increment the counters...
			done = increase(nu,0);
		}
	}
	delete[] nu;

	if (report_progress)
	{
		std::cerr<<(char)13;
		std::cerr<<"progress: 100%                        ";
		std::cerr.flush();
	}
	
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::update_h_best_list(const float& wght, int* nu, w_bit_set& intersection)
// ==============================================================
{
	
	int eta,i;
	float worst_overlapping_wght = FLT_MAX;
	float worst_non_overlapping_wght = FLT_MAX;
	int worst_overlapping_idx = -1;
	int worst_non_overlapping_idx = -1;
	
	if (h_so_far<h)
	{
		anything_found = true;
		// we didnt even have h many solutions so far... so
		// theres space anyway to fill with one more solution
		
		// but first, check whether the new solution overlaps with an exisiting one...
		bool replaced = false;
		for (eta=0; eta<h_so_far; eta++)
			if (!differs(nu,best_h_nu[eta]) && wght>best_h_wghts[eta])
			{
				for (int i=0; i<K; i++)
					best_h_nu[eta][i] = nu[i];
				best_h_wghts[eta] = wght;
				best_h_intersections[eta] = intersection;
				replaced = true;
			}
		if (!replaced)
		{
			for (i=0; i<K; i++)
				best_h_nu[h_so_far][i] = nu[i];
			best_h_wghts[h_so_far] = wght;
			best_h_intersections[h_so_far] = intersection;
			h_so_far++;
		}
		if (wght>h_best_weight)
			h_best_weight = wght;
		return;
	}
	
	if (wght>h_best_weight)
	{
		// if we reach this point compute_h_best_intersection(), we have
		//
		//    intersection.cardinality() > h_best_weight)
		//
		// where intersection is the intersection of all
		// C[i][nu[i]], for all i in [1:K]. In other words,
		// this means that we have found a new intersection
		// that might be one the h best intersections. Whether
		// this is actually the case, or whether we already
		// know a very similar, but better intersection is to
		// be figured out next.

		// I. determine the worst of the saved solutions that
		// overlaps with the new candidate solution and the worst
		// of the saved solutions that does not overlap with the
		// new candidate solution
					
		for (eta=0; eta<h; eta++)
		{
			if (!differs(nu,best_h_nu[eta]))
			{
				if (best_h_wghts[eta]<worst_overlapping_wght)
				{
					worst_overlapping_wght = best_h_wghts[eta];
					worst_overlapping_idx = eta;
				} 
			} else {
				if (best_h_wghts[eta]<worst_non_overlapping_wght)
				{
					worst_non_overlapping_wght = best_h_wghts[eta];
					worst_non_overlapping_idx = eta;
				} 
			}
		}
					
				
		// IIa. If the worst overlapping solution is worse than the
		// new candidate solution, throw away the old solution by the
		// new one.
					
		if (worst_overlapping_wght<wght)
		{
			for (int i=0; i<K; i++)
				best_h_nu[worst_overlapping_idx][i] = nu[i];
			best_h_wghts[worst_overlapping_idx] = wght;
			best_h_intersections[worst_overlapping_idx] = intersection;
		}
					
		// III. If there was no solution that overlaps with the
		// new candidate solution, then replace the worst solution
		// by the new one.
		if (worst_overlapping_idx==-1)
		{
			for (int i=0; i<K; i++)
				best_h_nu[worst_non_overlapping_idx][i] = nu[i];
			best_h_wghts[worst_non_overlapping_idx] = wght;
			best_h_intersections[worst_non_overlapping_idx] = intersection;
		}
					
		// Note: If the worst overlapping solution is better than the
		// new candidate solution, throw away the new solution.

		// now, recompute h_best_weight
		h_best_weight = FLT_MAX;
		for (eta=0; eta<h; eta++)
			if (best_h_wghts[eta]<h_best_weight)
				h_best_weight = best_h_wghts[eta];
	}
	
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::compute_h_best_intersections()
// ==============================================================
{
	int i, mu, eta;
	int C_size = 1;
	int* nu = new int[K];
	//best_h_wghts = new float[h];

	// we dont deal with thresholding, so we can set the h-th best weight to 0.
	h_best_weight = 0.;
	
	// the number of solutions found so far is nadazero.
	h_so_far = 0;

	//best_h_nu = new int*[h];
	//for (eta=0; eta<h; eta++)
	//	best_h_nu[eta] = new int[K];
	
	for (i=0; i<K; i++)
	{
		nu[i] = 0;
		if (Lambda[i]==0)
		{
			anything_found = false;
			return;
		}
	}

	if (report_progress)
	{
		std::cerr.precision(32);		
		progress = 0;
		prgr_count = 0;
		num_of_inst = 1;
		for (i=0; i<K; i++)
			num_of_inst *= (double)(Lambda[i]);
		std::cerr<<(char)13;
		std::cerr<<"progress: 0%                        ";
		std::cerr.flush();
	}
	num_of_inst--;
	
	anything_found = true;	
	bool done = false;
	while (!done)
	{
		if (report_progress)
			if (prgr_count==65536)
			{
				mpf_class x(num_of_inst);
				mpf_class y(num_of_inst);
				mpf_class z(num_of_inst);
				// compute number of visited instances Horner-like
				x = nu[0];
				for (i=1; i<K; i++)
				{
					x *= (Lambda[i]);
					x += (nu[i]);
				}
				z = x/y;
				z *= mpf_class(100.);
				std::cerr<<(char)13;
				std::cerr<<"progress: "<<z<<"%                        ";
				std::cerr.flush();
				prgr_count = 0;
			} else {
				prgr_count++;
			}
		// compute the intersection...
		w_bit_set intersection = C[0][nu[0]];
		bool bound = false;
		float wght;
		for (i=1; i<K; i++)
		{
			intersection.intersect(C[i][nu[i]]);
			// The following bound step is based on the premise that 1.0
			// is an upper bound for the weight of a match, i.e., an
			// exact occurence of s_j in T_i is weighted by 1.0 and all
			// approximate occurence are weighted lower than 1.0.
			wght = intersection.weight();
			if (wght+(float)(K-i+1)<h_best_weight)
			{
				// perform a bound-step in the branch-and-bound-algorithm
				done = increase(nu,i);
				bound = true;
				break;
			}
		}

		if (!bound)
		{
			update_h_best_list(wght,nu,intersection);
			// finally increment the counters...
			done = increase(nu,0);
		}
	}
	delete[] nu;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_element_array(w_bit_set M, int* E, int& l)
// ==============================================================
{
	int j;
	l = 0;
	for (int j=0; j<k; j++)
		if (M.contains(j))
		{
			E[l] = j;
			l++;
		}
}


// ==============================================================
template<class w_bit_set>
w_bit_set weighted_footprint_detector<w_bit_set>::difference(w_bit_set M, w_bit_set& D)
// ==============================================================
{
	int nu;
	for (nu=0; nu<k; nu++)
	{
		if (M.contains(nu) && D.contains(nu))
			M.remove(nu,0.);
	}
	return M;
}


// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::test_diff_set(w_bit_set& M)
// ==============================================================
{
	int i;
	std::cout<<"searching ";
	print_with_names(M);
	std::cout<<"="<<M<<"\n";
	bool found;
	bool all_found = true;
	float best_wght = 0.;
	
	for (i=0; i<K; i++)
	{
		found = false;
		float best_wght_i = 0.;
		arrangement_range hits = A[i].equal_range(M);
		typename arrangement::iterator a_it = hits.first;
		typename arrangement::iterator a_end = hits.second;		
		for (; a_it!=a_end; a_it++)
		{
			w_bit_set A_i_j = *a_it;
			float wght = A_i_j.weight();
			if (wght > best_wght_i)
				best_wght_i = wght;
			found = true;
		}
		all_found = all_found & found;
		best_wght += best_wght_i;
	}
	if (all_found && !weighted)
		best_wght = M.cardinality();
	
	if (all_found && best_wght>=overall_best_weight)
	{
		std::cout<<"new sol. w/ wght "<<best_wght<<"\n";
		overall_best_weight = best_wght;
		best_intersection = M;
		// best_nu is unknown at this point and will be reconstructed
		// after the whole optimization procedure.
		return true;
	}
	return false;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::enumerate_differences(int D)
// ==============================================================
{
	int i,j,nu,q,d,kk;
	int* c = new int[D+1];
	int* E = new int[k];
	// the arrangement multisets A are a class member variable so that
	// we dont need to pass them as a parameter.
	A = new arrangement[K];
	cell_set all_cells;
	w_bit_set S;

	// fill arrangement A and cell set all_cells with the corresponding
	// entries from C
	for (i=0; i<K; i++)
	{
		arrangement* A_i = &(A[i]);
		int Lambda_i = Lambda[i];
		w_bit_set* C_i = C[i];
		for (nu=0; nu<Lambda_i; nu++)
		{
			w_bit_set* C_i_nu = &(C_i[nu]);
			//A_i->insert(*C_i_nu);
			//all_cells.insert(*C_i_nu);


			// not implemented as a member of w_bit_sit in order keep the
			// w_bit_set interface simple.
			get_element_array(*C_i_nu,E,kk);
			if (kk<D)
				D = kk;
		
			// enumerate all D-element subsets of [1:k] using the revolving
			// door method given as Algorithm R in Knuth 7.2.1.3 (Generating
			// all Combinations). Implementation as an FSA with 4 states
			// corresponding to steps R2,...,R5. The advantage of using the
			// revolving door method is that the number of operations on the
			// set S is minimized (since we only perform 2 operations in
			// each enumeration step).
			for (j=0; j<D; j++)
			{
				c[j] = j;
				S.set(E[j],0.);
			}
			c[D] = kk;
			int state = 2;
			j = 0;
			if (D>0)
				while (j<=D)
				{
					switch (state)
					{
						case 2: {
							w_bit_set M = difference(*C_i_nu,S);
							//std::cout<<"ins card "<<M.cardinality()<<" weight "<<M.weight()<<"\n";
							A_i->insert(M);
							all_cells.insert(M);
							state = 3;
							break;
						}
						case 3:
							if (D%2==1)
							{
								// D odd
								if (c[0]+1<c[1])
								{
									S.remove(E[c[0]],0.);
									c[0]++;
									S.set(E[c[0]],0.);
									state = 2;
								} else {
									j = 2;
									state = 4;						
								}
							} else {
								// D even
								if (c[0]>0) {
									S.remove(E[c[0]],0.);
									c[0]--;
									S.set(E[c[0]],0.);
									state = 2;
								} else {
									j = 2;
									state = 5;
								}	
							}
							break;
						case 4:
							if (c[j-1]>=j) {
								S.remove(E[c[j-2]],0.);
								c[j-1] = c[j-2];
								c[j-2] = j-2;
								S.set(E[j-2],0.);
								state = 2;
							} else {
								j++;
								state = 5;
							}
							break;
						case 5:
							if (c[j-1]+1<c[j]) {
								S.remove(E[c[j-2]],0.);
								c[j-2] = c[j-1];
								c[j-1]++;
								S.set(E[c[j-1]],0.);
								state = 2;
							} else {
								j++;
								state = 4;
							}
							break;
					}
				}
			else {
				// D=0 means we dont need to subtract any members from *C_i_nu...
				A_i->insert(*C_i_nu);
				all_cells.insert(*C_i_nu);
			}
		}
		//std::cout<<"insterted "<<A_i->size()<<" elms; should be "<<Lambda_i<<".\n";
	}

	bool new_best = false;
	typename cell_set::iterator c_it = all_cells.begin();
	typename cell_set::iterator c_end = all_cells.end();
	for (; c_it != c_end; c_it++)
	{
		w_bit_set M = *c_it;
		new_best = new_best || test_diff_set(M);
	}

	anything_found = anything_found || new_best;
	delete[] A;
	// it remains to compute best_nu
	if (new_best)
		for (i=0; i<K; i++)
		{
			float best_i = 0.;
			for (j=0; j<Lambda[i]; j++)
			{
				if (best_intersection.subseteq(C[i][j]) && C[i][j].weight()>best_i)
				{
					best_i = C[i][j].weight();
					best_nu[i] = j;
				}
			}
		}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_frag_occurences(int i, int j, match_list& match_pos)
// ==============================================================
{

	int ii,jj;

	// allocate memory for the distance matrix
	// int** D = new int*[n[j]+m[j]];
	// for (ii=0; ii<n[j]+M[j], ii++)
	//	D[ii] = new int[n[j]];
	int* last_row = new int[N[i]+m[j]];
	int* last_row_where_from = new int[N[i]+m[j]];
	int* curr_where_from =  new int[n[j]];
	int* prev_where_from =  new int[n[j]];
	
	int* curr_col = new int[n[j]];
	int* prev_col = new int[n[j]];
	curr_col[0] = (T[i][0]==s[j][0]) ? 0 : 1;
	curr_where_from[0] = 0;
	for (jj=1; jj<n[j]; jj++)
	{
		curr_col[jj]=curr_col[jj-1]+1;
		curr_where_from[jj] = 0;
	}
	last_row[0] = curr_col[n[j]-1];
	last_row_where_from[0] = 0;

	int N_i = N[i];
	
	for (ii = 1; ii<N_i; ii++)
	{
		int* buf;
		
		// compute alignment's matrix column ii
		
		buf = prev_col;
		prev_col = curr_col;		
		curr_col = buf;
		// same thing with curr_where_from and prev_where_from.
		buf = prev_where_from;		
		prev_where_from = curr_where_from;
		curr_where_from = buf;
		curr_col[0] = (T[i][ii]==s[j][0] ? 0 : 1);
		curr_where_from[0] = (T[i][ii]==s[j][0] ? ii : prev_where_from[0]);
		int n_j = n[j];
		for (jj=1; jj<n_j; jj++)
		{
			// This is the dynamic programming recurrence (somewhat
			// difficult to recongnize as such due to memory saving
			// implementation...)
			int m_cost = prev_col[jj-1] + (T[i][ii]==s[j][jj] ? 0 : 1);

			/// ATTENTION: DYNAMIC PROGRAMMING HAS BEEN DISABLED IN THE
			/// FOLLOWING TWO LINES!  MISMATCHES INSTEAD OF EDIT-DISTANCE
			/// ARE USED; TO SWITCH BACK TO EDIT DISTANCE, REPLACE
			/// INT_MAX BY THE COMMENTED CODE.
			int d_cost = INT_MAX;//curr_col[jj-1] + 1;
			int i_cost = INT_MAX;//prev_col[jj]   + 1;
			if (m_cost<=d_cost)
				if (m_cost<=i_cost)
				{
					curr_col[jj] = m_cost;
					curr_where_from[jj] = prev_where_from[jj-1];
				} else {
					curr_col[jj] = i_cost;
					curr_where_from[jj] = prev_where_from[jj];
				}
			else if (i_cost<=d_cost)
			{
				curr_col[jj] = i_cost;
				curr_where_from[jj] = prev_where_from[jj];
			} else {
				curr_col[jj] = d_cost;
				curr_where_from[jj] = curr_where_from[jj-1];
			}
		}
		last_row[ii] = curr_col[n[j]-1];
		last_row_where_from[ii] = curr_where_from[n[j]-1];		
	}

	// We only store those occurences that are not very similar to
	// other, definitely better occurences with a lower edit
	// distance. I.e., we only take into account those alignment paths
	// that end in a local minimum in the last row.
	for (ii=n[j]-1; ii<N_i; ii++)
		if (last_row[ii]<=m[j] &&
			last_row[ii-1]>=last_row[ii] &&
			last_row[ii+1]>=last_row[ii])
		{
			//std::cout<<"Found pattern "<<j<<" in text "<<i
			//			<<" at position "<<last_row_where_from[ii]
			//			<<" with "<<last_row[ii]<<" mismatches;"
			//			<<" path ends in "<<ii<<".\n";					
			match_pos.push_back(std::triple<int,int,bool>(last_row_where_from[ii],last_row[ii],!(reverse && j>=kappa/2)));
		}
	
	delete[] curr_col;
	delete[] prev_col;
	delete[] last_row;
	delete[] curr_where_from;
	delete[] prev_where_from;
	delete[] last_row_where_from;
	
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_occurences(int i, int j)
// ==============================================================
{
	
	match_list match_pos;
	get_frag_occurences(i,j,match_pos);
	
	int LL = match_pos.size();
	lambda[i][j] = LL;
	p[i][j] = NULL;
	M[i][j] = NULL;
	R[i][j] = NULL;
	int* x = new int[LL];
	p[i][j] = x;
	M[i][j] = new int[LL];
	R[i][j] = new bool[LL];

	match_list::iterator m_it = match_pos.begin();
	match_list::iterator m_end = match_pos.end();;

	int nu = 0;
	for (; m_it!=m_end; ++m_it, ++nu)
	{
		p[i][j][nu] = m_it->first;
		M[i][j][nu] = m_it->second;
		R[i][j][nu] = m_it->third;
	}
	
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_group_occurences(int i, int j)
// ==============================================================
{

	// match list for one single fragment belonging to group j
	match_list match_pos;

	// match list for the whole group j, which is merged from all the
	// group's fragments' match lists.
	match_list merged_match_pos;
	
	for (int iota = 0; iota<group_sizes[j]; iota++)
	{
		match_pos.clear();
		int j_ = group_indices[j][iota];
		get_frag_occurences(i,j_,match_pos);
		merged_match_pos.merge(match_pos);
	}
	
	int LL = merged_match_pos.size();
	lambda[i][j] = LL;
	p[i][j] = NULL;
	M[i][j] = NULL;
	R[i][j] = NULL;
	int* x = new int[LL];
	p[i][j] = x;
	M[i][j] = new int[LL];
	R[i][j] = new bool[LL];

	match_list::iterator m_it = merged_match_pos.begin();
	match_list::iterator m_end = merged_match_pos.end();;

	int nu = 0;
	for (; m_it!=m_end; ++m_it, ++nu)
	{
		p[i][j][nu] = m_it->first;
		M[i][j][nu] = m_it->second;
		R[i][j][nu] = m_it->third;
	}
}


/// compute group names form fragment names as well as the
/// corresponding mapping group_indices that assigns a set of fragment
/// index to each group index.
// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_groups()
// ==============================================================
{
	
	int j, iota;
	std::string group_name;
	std::set<std::string> groups;

	if (!grouped && reverse)
	{
		grouped = true;
		kappa = k;
		k /= 2;
		group_names = new char*[k];
		group_sizes = new int[k];
		group_indices = new int*[k];
		for (j=0; j<k; j++)
		{
			group_names[j] = new char[strlen(s_names[j])];
			strcpy(group_names[j],s_names[j]);
			group_sizes[j] = 2;
			group_indices[j] = new int[2];
			group_indices[j][0] = j;
			group_indices[j][1] = j+k;
		}
	} else {
		group_to_index_list g_idx_list;
		kappa = k;
		
		for (j=0; j<k; j++)
		{
			if (grouped)
				group_name = std::string(s_names[j],num_of_group_characters);
			else
				// the fragment name turns into a group name as well (note that s_names[j] is null-terminated)
				group_name = std::string(s_names[j]);
			groups.insert(group_name);
			g_idx_list[group_name].push_back(j);
		}
		
		k = groups.size();
		group_names = new char*[kappa];
		group_sizes = new int[kappa];
		group_indices = new int*[kappa];
		std::set<std::string>::iterator g_it = groups.begin();
		std::set<std::string>::iterator g_end = groups.end();
		for (j=0; g_it!=g_end; g_it++, j++)
		{
			std::string g_name = *g_it;
			group_names[j] = new char[g_name.size()+1];
			strcpy(group_names[j],g_name.c_str());
			group_sizes[j] = g_idx_list[g_name].size();
			group_indices[j] = new int[group_sizes[j]];
			std::list<int>::iterator i_it = g_idx_list[g_name].begin();
			std::list<int>::iterator i_end = g_idx_list[g_name].end();
			for (iota = 0; i_it!=i_end; i_it++, iota++)
				group_indices[j][iota] = *i_it;	
		}
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::run(int L_, float threshold_)
// ==============================================================
{
	
	int i, j, ii, jj, eta;

	L = L_;
	threshold = threshold_;
	
	p = new int**[K];
	M = new int**[K];
	R = new bool**[K];
	lambda =  new int*[K];
	Lambda = new int[K];
	P = new int*[K];
	C = new w_bit_set*[K];

	best_nu = new int[K];
	last_nu = new int[K];

	num_of_solutions = 0;
	
	if (h>1)
	{
		best_h_nu = new int*[h];
		best_h_wghts = new float[h];
		best_h_intersections = new w_bit_set[h];
		for (eta=0; eta<h; eta++)
			best_h_nu[eta] = new int[K];		
	} else {
		best_h_nu = NULL;
		best_h_wghts = NULL;
	}

	// split fragments into groups accordind to the fragment names'
	// prefixes, if grouped==true; otherwise, assign one group to each
	// fragment name.
	get_groups();

	// compute nucleotide- and dinucleotide frequencies for eventually
	// used weighting schemes that rely on these data.
	get_freq();

	// prepare C[i][j], P[i][j] etc. by performing a sweep algorithm
	init_arrangement_cells();

	if (enum_differences) {
		int d;
		overall_best_weight = 0.;
		for (d=0; d<=delta; d++)
			enumerate_differences(d);
	} else if (algorithm==1) {
		std::cerr<<"computing intersections...\n";
		// apply Algorithm (A1)
		if (h<=1)
		{
			compute_intersections();
		} else {
			compute_h_best_intersections();
			//print_h_results();
		}
	} else {
		// algorithm==2
		// apply Algorithm (A2)
		std::cerr<<"enumerating subsets...\n";
		if (h<=1)
			enumerate_subsets();
		else
			enumerate_h_best_subsets();
	}
	if (threshold>=0.)
		std::cerr<<num_of_solutions<<" matches were found.\n";

	if (ps_output)
		print_ps_labels(std::string("labels.ps"));
}


// ==============================================================
template<class w_bit_set>
bool weighted_footprint_detector<w_bit_set>::differs(int* nu_1, int* nu_2)
// ==============================================================
{
	// difference policy: two matches are different if in at least one
	// sequence the corresponding intervals of length L do not
	// intersect.
	int i;
	for (i=0; i<K; i++)
		if (nu_1[i]-nu_2[i]>=L)
			return true;
	return false;
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_with_names(w_bit_set& B)
// ==============================================================
{
	int count = B.cardinality();
	std::cout<<"{";
	int printed = 0;
	for (int j=0; j<k; j++)
		if (B.contains(j))
		{
			if (grouped)
				std::cout<<group_names[j];
			else
				std::cout<<s_names[j];
			if (printed < count-1)
				std::cout<<",";
			printed++;
		}
	std::cout<<"}";
}


// ==============================================================
template<class w_bit_set>
float weighted_footprint_detector<w_bit_set>::get_result()
// ==============================================================
{
	return overall_best_weight;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_h_results()
// ==============================================================
{
	for (int eta=0; eta<h_so_far; eta++) 
		print_single_result(best_h_nu[eta], best_h_intersections[eta], best_h_wghts[eta]);
}

	
// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_ps_h_results(std::string& filename)
// ==============================================================
{

	for (int eta=0; eta<h_so_far; eta++)
	{
		std::string fname = filename;
		std::filebuf fb;
		char buf[50];
		sprintf(buf, "%d", eta);
		if (fname.length()>3)
			if (fname[fname.length()-3]=='.' && fname[fname.length()-2]=='p' &&fname[fname.length()-1]=='s')
				fname.erase(fname.end()-3,fname.end());
		fname.append(std::string(buf));
		fname.append(std::string(".ps"));
		fb.open (fname.c_str(),std::ios::out);
		std::ostream os(&fb);
		print_ps_header(os);
		print_ps_single_result(best_h_nu[eta],os,best_h_intersections[eta]);
		print_ps_footer(os);
		fb.close();
	}
}

	
// ==============================================================
template<class w_bit_set>
float weighted_footprint_detector<w_bit_set>::print_result()
// ==============================================================
{

	if (count_results)
		return 0.;
	
	if (!anything_found)
	{
		std::cout<<"No L-occurences found.\n";
		return 0.;
	}

	if (ps_output)
		print_postscript(ps_filename);
	
	if (h>1)
	{
		print_h_results();
		return 0.;
	}
	print_single_result(best_nu, best_intersection,overall_best_weight);
	return overall_best_weight;
	
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_ps_labels(std::string filename)
// ==============================================================
{
	int j, last_k;

	
	std::filebuf fb;
	fb.open (filename.c_str(),std::ios::out);
	std::ostream os(&fb);
	print_ps_header(os);
	
	// compute max. string length of a group- or fragment name in order
	// to determine a suitable gap length
	int max_str_len = 0;
	for (int j=0; j<k; j++)
	{
		if (grouped)
		{
			if (strlen(group_names[j])>max_str_len)
			  max_str_len = strlen(group_names[j]);
		} else {
			if (strlen(s_names[j])>max_str_len)
				max_str_len = strlen(s_names[j]);
		}
	}
	
	int x = 0;
	int jj;
	
  for (int j=0, jj=0; j<k; j++, jj++)
  {
	  if (jj==60)
	  {
		  jj = 0;
		  x += (30 + 9*max_str_len);
	  }
	  int y = 15*jj;
	  if (grouped)
	  {
		  print_label(x,y,j,group_names[j],os);
	  } else {
		  print_label(x,y,j,s_names[j],os);
	  }
  }
  print_ps_footer(os);
  fb.close();
  
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_postscript(std::string filename)
// ==============================================================
{

	if (!anything_found)
		return;
	
  std::filebuf fb;
  if (threshold<0.)
  {
	  if (filename.length()>3)
	  {
		  if (filename[filename.length()-3]!='.' && filename[filename.length()-2]!='p' &&filename[filename.length()-1]!='s')
			  filename.append(std::string(".ps"));
	  } else {
		  filename.append(std::string(".ps"));		  
	  }
	  fb.open (filename.c_str(),std::ios::out);
  } else {
	  char buf[50];
	  sprintf(buf, "%d", num_of_solutions);
	  if (filename.length()>3)
		  if (filename[filename.length()-3]=='.' && filename[filename.length()-2]=='p' &&filename[filename.length()-1]=='s')
			  filename.erase(filename.end()-3,filename.end());
	  filename.append(std::string(buf));
	  filename.append(std::string(".ps"));
	  fb.open (filename.c_str(),std::ios::out);
  }
  std::ostream os(&fb);
  if (h>1)
  {
	  print_ps_h_results(filename);
	  return;
  }
  
  print_ps_header(os);
  
  print_ps_single_result(best_nu,os,best_intersection);
  print_ps_footer(os);
  fb.close();
	
}

// ==============================================================
template<class w_bit_set>
float weighted_footprint_detector<w_bit_set>::print_single_result(int* nu, w_bit_set& intersection, const float& weight)
// ==============================================================
{

	std::string weightsize;
	std::string rev_string;
	if (reverted)
		rev_string = "reverted ";
	else
		rev_string = "";
	
	if (weighted)
		weightsize="weight ";
	else
		weightsize="size ";
	if (threshold<0.)
		std::cout<<"The Max. "<<rev_string<<"L-Occurence has "<<weightsize<<weight<<";\n";
	else
		std::cout<<"\nFound an "<<rev_string<<"L-Occurence with "<<weightsize<<weight<<";\n";
	std::cout<<"contains fragments ";
	print_with_names(intersection);
	std::cout<<".\n";
	for (int ii=0; ii<K; ii++)
	{
		if (output_detail)
			std::cout<<"L-Occ. in ";
		std::cout<<T_names[ii];
		if (output_detail)
			std::cout<<" ranges from pos.";
		std::cout<<" "<<max(0,P[ii][nu[ii]]-L);
		if (output_detail)
			std::cout<<" to pos.";
		std::cout<<" "<<P[ii][nu[ii]]<<".\n";
	}

	// now we figure out the actual positions of each s_j in each T_i
	// in the best L-occurence.
	int i,j,mu;
	
	for (i=0; i<K; i++)
	{
		if (output_detail)
			std::cout<<"L-OCCURENCE IN ";
		std::cout<<T_names[i]<<":\n";
		if (output_detail)
			std::cout<<"Fragments occuring in the "<<rev_string<<"L-occ. are ";
		print_with_names(C[i][nu[i]]);
		if (output_detail)
			std::cout<<".";
		std::cout<<std::endl;
		cluster_set cluster_names;
		for (j=0; j<k; j++)
		{
			// we collect all matches in a list of 
			if (intersection.contains(j))
			{
				int pos_i_j;
				float best_weight = 0.;
				int best_dist = 0;
				int lambda_i_j = lambda[i][j];
				for (mu=0; mu<lambda_i_j; mu++)
				{
					int curr_p = p[i][j][mu];
					float curr_w = get_weight(i,j,mu);
					if (curr_p>=P[i][nu[i]]-L && curr_p<=P[i][nu[i]] && curr_w>=best_weight)
					{
						pos_i_j = curr_p;
						best_weight = curr_w;
						best_dist = M[i][j][mu];
					}
				}
				cluster_names.insert(cluster_triple(pos_i_j,best_dist,j));
			}
		}
		cluster_set::iterator c_it  = cluster_names.begin();
		cluster_set::iterator c_end = cluster_names.end();
		for (; c_it!=c_end; c_it++)
		{
			int pos = c_it->first;
			int dist = c_it->second;
			j = c_it->third;
			std::cout<<group_names[j]<<" ";
			if (output_detail)
				std::cout<<"with length ";
			std::cout<<n[j]<<" ";
			if (output_detail)
				std::cout<<"at pos. ";
			std::cout<<pos<<" ";
			if (output_detail)
				std::cout<<"with ";					
			float curr_w = ((float)(m[j]-dist)+1.) / (((float)(m[j])+1.));
			if (weighted)
			{
				if (output_detail)
					std::cout<<"weight ";
				std::cout<<curr_w<<" ";
				if (output_detail)
					std::cout<<"and ";
			}
			if (output_detail)
				std::cout<<"edit dist. ";
			std::cout<<dist<<"\n";
		}
	}
	std::cout<<"------------------------------------------------------"<<std::endl;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::print_ps_single_result(int* nu, std::ostream& o, w_bit_set& intersection)
// ==============================================================
{

	// figure out the actual positions of each s_j in each T_i
	// in the best L-occurence.
	int i,j,mu;

	print_cluster_members<w_bit_set>(50+70*K,intersection,o);
	for (i=0; i<K; i++)
	{
		ps_match_list ml;
		for (j=0; j<k; j++)
		{
			//if (intersection.contains(j))
			//{
			int pos_i_j;
			float best_weight = 0.;
			int best_dist = 0;
			int lambda_i_j = lambda[i][j];
			for (mu=0; mu<lambda_i_j; mu++)
			{
				int curr_p = p[i][j][mu];
				float curr_w = get_weight(i,j,mu);//((float)(m[j]-M[i][j][mu])+1.) / (((float)(m[j])+1.));
				if (curr_p>=P[i][nu[i]]-L && curr_p<=P[i][nu[i]])
				{
					/*if (!grouped && reverse && (j>=k/2))
						ml.push_back(match_list_triple(j-k/2,curr_p,R[i][j][mu]));
						else*/
					ml.push_back(match_list_triple(j,curr_p,R[i][j][mu]));						
				}
			}
			//}
		}
		print_cluster(50+70*(K-i-1), T_names[i], max(0,P[i][nu[i]]-L), P[i][nu[i]], L, ml, o);
	}
}

/// compute the reverse complement of the patterns, so that in effect
//the genome is searched in backward direction for L-occurences.
//==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::reverse_compl()
// ==============================================================
{

	reverted = true;
	
	int i;
	
	char* compl_table = new char[256];

	for(i=0; i<256; i++) compl_table[i]=(int)'*';
	
	compl_table[(int)'A']='T';
	compl_table[(int)'C']='G';
	compl_table[(int)'G']='C';
	compl_table[(int)'T']='A';
	compl_table[(int)'R']='Y';
	compl_table[(int)'Y']='R';
	compl_table[(int)'M']='K';
	compl_table[(int)'K']='M';
	compl_table[(int)'S']='S';
	compl_table[(int)'W']='W';
	compl_table[(int)'H']='D';
	compl_table[(int)'B']='V';
	compl_table[(int)'V']='B';
	compl_table[(int)'D']='H';
	compl_table[(int)'N']='N';
	compl_table[(int)'a']='t';
	compl_table[(int)'c']='g';
	compl_table[(int)'g']='c';
	compl_table[(int)'t']='a';
	compl_table[(int)'r']='y';
	compl_table[(int)'y']='r';
	compl_table[(int)'m']='k';
	compl_table[(int)'k']='m';
	compl_table[(int)'s']='s';
	compl_table[(int)'w']='w';
	compl_table[(int)'h']='d';
	compl_table[(int)'b']='v';
	compl_table[(int)'v']='b';
	compl_table[(int)'d']='h';
	compl_table[(int)'n']='n';
	
	int j, iota;
	for (j=0; j<k; j++)
	{
		int n_j = n[j];
		int half = n_j;
		for (iota=0; iota<half; iota++)
		{
			char tmp = compl_table[s[j][iota]];
			s[j][iota] = compl_table[s[j][n_j-iota]];
			s[j][n_j-iota] = tmp;
		}
		if (half%2==1)
		{
			s[j][half+1] = compl_table[s[j][half+1]] ;
		}
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_grouping(bool grp)
// ==============================================================
{
	grouped = grp;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_reverse(bool rev)
// ==============================================================
{
	reverse = rev;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_postscript(bool ps)
// ==============================================================
{
	ps_output = ps;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_ps_filename(std::string psf)
// ==============================================================
{
	ps_filename = psf;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_progress_report(bool rpt)
// ==============================================================
{
	report_progress = rpt;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_algorithm(int A)
// ==============================================================
{
	if (A==1 || A==2)
		algorithm = A;
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_output_detail(bool od)
// ==============================================================
{
	output_detail = od;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_number_of_hits(int h_)
// ==============================================================
{
	h = h_;
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_weighted(bool wght)
// ==============================================================
{
	weighted = wght;
	w_bit_set::weighted = weighted;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_count_results(bool cr)
// ==============================================================
{
	count_results = cr;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_weight_type(int wt)
// ==============================================================
{
	weight_type = wt;
}


// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_group_characters(int n_o_g_c)
// ==============================================================
{
	num_of_group_characters = n_o_g_c;
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::set_delta(int delta_)
// ==============================================================
{
	enum_differences = true;
	delta = delta_;
}



// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::clear()
// ==============================================================
{
	int i,j,nu;
	if (p!=NULL)
	{
		for (i = 0; i<K; i++)
		{
			for (j = 0; j<k; j++)
				delete[] p[i][j];
			delete[] p[i];
		}
		delete[] p;
		p = NULL;
	}	
	if (M!=NULL)
	{
		for (i = 0; i<K; i++)
		{
			for (j = 0; j<k; j++)
				delete[] M[i][j];
			delete[] M[i];
		}
		delete[] M;
		M = NULL;
	}
	if (C!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] C[i];
		delete[] C;
		C = NULL;
	}
	if (P!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] P[i];
		delete[] P;
		P = NULL;
	}
	if (lambda!=NULL)
	{
		for (i = 0; i<K; i++)
			delete[] lambda[i];
		delete[] lambda;
		lambda = NULL;
	}
	if (Lambda!=NULL)
	{
		delete[] Lambda;
		Lambda = NULL;
	}

	if (best_nu!=NULL)
	{
		delete[] best_nu;
		best_nu = NULL;
	}

	if (last_nu!=NULL)
	{
		delete[] last_nu;
		last_nu = NULL;
	}

	// TODO: delete everything that has to do with grouping.
	
}

// ==============================================================
template<class w_bit_set>
weighted_footprint_detector<w_bit_set>::~weighted_footprint_detector()
// ==============================================================
{
	clear();
}

// ==============================================================
template<class w_bit_set>
float weighted_footprint_detector<w_bit_set>::get_weight(int i, int j, int nu)
// ==============================================================
{
	switch (weight_type)
	{
		case 0:
			return 1.;
			break;
		case 1:
			return ((float)(m[j]-M[i][j][nu])+1.) / (((float)(m[j])+1.));
			break;
		case 2:
			return -log(p_value(i,j,nu));
			std::cerr<<"log p_value: "<<-log(p_value(i,j,nu))<<"\n";
		default:
			return 1.;
	}
}


// ==============================================================
template<class w_bit_set>
double weighted_footprint_detector<w_bit_set>::p_value(int i, int j, int nu)
// ==============================================================
{

	int jj, kk;
	int pos = p[i][j][nu];
	double seq_N = (double)N[i];
	double p = 1.;
	if (s[j][0]==T[i][pos])
		p *= (double)frequencies[i][index_table[s[j][0]]]/seq_N;
	int kk_end = n[i]-1;
	for (kk=0, jj=pos; kk<kk_end; jj++, kk++)
		if (s[j][kk]==T[i][jj])
			if (s[j][kk+1]==T[i][jj+1])
			{
				double p_local = (double)di_frequencies[i][index_table[T[i][jj]]][index_table[T[i][jj+1]]]/seq_N;
					p *= p_local;
			} else {
				double p_local = (double)frequencies[i][index_table[T[i][jj+1]]]/seq_N;					
				p *= p_local;
			}
	return p;
	
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::get_freq()
// ==============================================================
{
	int a,i,j;
	di_frequencies = new int**[32];
	frequencies = new int*[32];

	for (a=0; a<K; a++)
	{
		frequencies[a] = new int[32];
		di_frequencies[a] = new int*[32];
		int* frq = frequencies[a];
		int** di_frq = di_frequencies[a];
		char* txt = T[a];
		
		for (i=0; i<32; i++)
		{
			/* we initialize all absolute frequencies with 4 and all
				dinucl.-freq with 1, so that we do not encounter 0
				probabilities, which may yield unwanted 0 probabilities for
				p-values. This corresponds to pretending the presence of a
				CCGGTTAACAGCTGATC in the genome sequence) */
			frq[i] = 4;
			di_frq[i] = new int[32];
			for (j=0; j<32; j++)
				di_frq[i][j] = 1;
		}

		int nn = N[a];
		for (i=0; i<nn-1; i++)
		{
			if ( (txt[i]=='C' || txt[i]=='G' || txt[i]=='T' || txt[i]=='A'
					  || txt[i]=='c' || txt[i]=='g' || txt[i]=='t' || txt[i]=='a') && 
				(txt[i+1]=='C' || txt[i+1]=='G' || txt[i+1]=='T' || txt[i+1]=='A'
				  || txt[i+1]=='c' || txt[i+1]=='g' || txt[i+1]=='t' || txt[i+1]=='a') )
			{
				di_frq[index_table[txt[i]]][index_table[txt[i+1]]]++;
				frq[index_table[txt[i]]]++;
			}
		}
	}
}

// ==============================================================
template<class w_bit_set>
void weighted_footprint_detector<w_bit_set>::compute_index_table()
// ==============================================================	
{

	int i;

	
	/// by default, we want to use asymmetric match tables, i.e.,
	/// wildcard Ns are only wildcards if they occur in a pattern
	/// fragment, but not in the text.
	bool asymmetric = true;

	index_table = new int[256];

	for(i=0; i<256; i++) index_table[i] = 30;
	
	index_table[65] = 0;   /* A */
	index_table[67] = 1;   /* C */
	index_table[71] = 2;   /* G */
	index_table[84] = 3;   /* T */
	index_table[85] = 3;   /* U == T */
	index_table[82] = 4;   /* R */
	index_table[89] = 5;   /* Y */
	index_table[77] = 6;   /* M */
	index_table[75] = 7;   /* K */
	index_table[83] = 8;   /* S */
	index_table[87] = 9;   /* W */
	index_table[72] = 10;  /* H */
	index_table[66] = 11;  /* B */
	index_table[86] = 12;  /* V */
	index_table[68] = 13;  /* D */
	if (asymmetric)
		index_table[78] = 14;  /* N */
	
	index_table[97] = 15;  /* a */
	index_table[99] = 16;  /* c */
	index_table[103] = 17; /* g  */
	index_table[116] = 18; /* t */
	index_table[117] = 18; /* u == t */
	index_table[114] = 19; /* r */
	index_table[121] = 20; /* y */
	index_table[109] = 21; /* m */
	index_table[107] = 22; /* k */
	index_table[115] = 23; /* s */
	index_table[119] = 24; /* w */
	index_table[104] = 25; /* h */
	index_table[98] = 26;  /* b */
	index_table[118] = 27; /* v */
	index_table[100] = 28; /* d */
	if (asymmetric)
		index_table[110] = 29; /* n */

}

} // end namespace bbq
