// bbq - Copyright (C) 2005, Axel Mosig, University of Leipzig. See
// main.cpp for details.

#ifndef WEIGHTED_FOOTPRINT_DETECTOR_H
#define WEIGHTED_FOOTPRINT_DETECTOR_H

#include <set>
#include <string>
#include <map>
#include <list>
#include <gmpxx.h>
#include "weighted_bitsets.h"
#include "weighted_bit_multi_set.h"
#include "bit_multi_set.h"
#include "triple.h"


//typedef uint_bitset bit_set;
//typedef uint64_bitset bit_set;
//typedef weighted_variable_bitset w_bit_set;



/** \c bbq is the namespace that contains all classes except for those
	 used for reading parameters etc. in main.cpp */
namespace bbq {

/// each list entry is coposed of a 
typedef std::list< std::triple<int,int,bool> > match_list;
typedef std::map<std::string,std::list<int> > group_to_index_list;

/// Class \c weighted_footprint_detector implements most crucial parts
/// of the algorithms underlying \c bbq, including computation of the
/// interval arrangements and finding best intersections between
/// cells. It takes a class \c w_bit_set that implements all set
/// operations needed by \c weighted_footprint_detector. Different
/// classes provided for \c w_bit_set implement weighted/unweighted
/// sets or multisets with different contraints on the cardinality of
/// the sets (and hence on the number of motifs that \c bbq can
/// handle).
template<class w_bit_set>
class weighted_footprint_detector
{
	
public:

	/// Procedure \c enumerate_differences needs fast finding of
	/// elements in the interval arrangements, which is achieved by
	/// this type of an stl \c multiset.
	typedef std::set< w_bit_set > arrangement;
	typedef std::pair< typename arrangement::iterator , typename arrangement::iterator > arrangement_range;
	typedef std::set< w_bit_set > cell_set;	
	
private:
	
	/// number of fragment sequences s_1,...,s_k.  Note that in case of
	/// grouped alignments (i.e., if grouped is true), k is the number
	/// of groups rather than the number of fragments!
	int k;

	/// in case of grouped alignments, kappa is the number of fragments
	/// (while k, the number of features, is taken by the number of
	/// groups.)
	int kappa;

	/// number of genome (excerpts) T_1,...,T_K
	int K;

	/// fragment sequences s_1,...,s_k or s_1,...,s_kappa in the
	/// grouped case.
	char** s;

	/// The number of leading characters of a fragment name that
	/// specifies the group the fragment belongs to. If grouped==false,
	/// all characters of a fragment name are interpreted as a group
	/// name -- two fragments belong to the same group iff they have
	/// the same name. This corresponds to the case
	/// num_of_group_characters==infty.
	int num_of_group_characters;

	/// sequences s_1,...,s_k (if grouped==true, otherwise without a
	/// meaning.)
	///
	/// This variable is initialized in the procedure get_groups.
	char** group_names;

	/// the group named group_names[j] consists of group_sizes[j] many
	/// fragments.
	///
	/// This variable is initialized in the procedure get_groups.
	int* group_sizes;

	/// the group named group_indices[j] consists of group_sizes[j]
	/// many indices, namely the indices of the fragments in the array
	/// s that the group consists of.
	///
	/// This variable is initialized in the procedure get_groups.
	int** group_indices;

	/// the group named group_names[j] consists of fragments
	///
	///   group_fragments[j][1],...,group_fragments[j][group_size[j]],
	///
	/// each of which is of type char*.
	///
	/// This variable is initialized in the procedure get_groups.
	//char*** group_fragments;
	
	/// names assigned to the fragment sequences s_1,...,s_k
	char** s_names;
	
	/// genome sequences T_1,...,T_K
	char** T;

	/// names assigned to the genome sequences T_1,...,T_k
	char** T_names;
	
	/// n_j contains the length of s_j
	int* n;

	/// N_i contains the length of T_i
	int* N;

	/// m_j contains the maximum edit distance accepted for an
	/// occurence of s_j.
	int* m;

	/// number of hits, i.e., the best h hits are computed. Hits are
	/// required to be significantly different. This is the case if
	/// differs(b1,b2) returns true, for two position vectors b1 and
	/// b2. Since there is no canonical notion of what "significantly
	/// different" means, a choice of several such difference-policies
	/// might be implemented in future releases.
	int h;

	/// since there might be less than h solutions, we count how many
	/// solutions have been found so far, in case less than h solutions
	/// have been considered. Otherwise, h_so_far == j.
	int h_so_far;

	/// For threshold<.0, only the best match is to be
	/// computed. Otherwise, all solutions whose weight (or
	/// cardinality, depending on the value of \c weighted) is at least
	/// \c threshold are computed. Depending on the value of \c
	/// count_results, the solutions are either counted or printed.
	float threshold;

	/// true if \c delta is to be used as an upper bound for the
	/// difference for solutions.
	bool enum_differences;

	/// upper bound for the number of fragments contained in a cell in
	/// addition to those that contibute to an optimal solution
	int delta;

	/// the solutions whose cardinality or weight (depending on the
	/// value of \c weighted) exceeding the threshold are counted (\c
	/// count_results=true) or printed printed. (\c count_results=false)
	bool count_results;

	/// whenever reverse() is called, reverted is toggled. Default
	/// value is false.
	bool reverted;

	/// Specifies the algorithm to be applied. Must be 1 or 2. A=1
	/// chooses (A1) (default), A=2 chooses (A2). The value can be
	/// changed using the \c set_algorithm() method.
	int algorithm;
	
	/// determines whether the fragments are grouped. If fragments are
	/// grouped, we are looking for L-occ. that contain a maximum
	/// number of groups rather than a maximum number of fragments. The
	/// group that a fragment belgons to is determined by the first 5
	/// characters of the fragment name.
	bool grouped;

	/// determines whether the output is to be made in a way that it
	/// can be understood by stupid users.
	bool output_detail;

	/// was there a successful match?
	bool anything_found;

	/// produce postscript output?
	bool ps_output;

	/// name of the file to write ps-output to, if \c ps_output=true.
	std::string ps_filename;
	
	/// number of instances inspected by Algorithm (A2), including
	/// particularly those that were bounded.
	mpz_class progress;

	int prgr_count;

	/// number of instances that Algorithm (A2) needs to inspect in
	/// total.
	mpz_class num_of_inst;

	/// p[i][j][nu] contains the position of the nu-th occurence of s_j
	/// in T_i.
	int *** p;

	/// R[i][j][nu] is true iff p[i][j][nu] is a reverse occurence of
	/// s_j in T_i.
	bool *** R;

	/// There are lambda[i][j] many occurences of s_j in T_i.
	int ** lambda;

	/// There are Lambda[i] many occurences of fragments in T_i.
	int * Lambda;

	/// C[i] contains a representation of the interval arrangement
	/// induced by T[i]. C[i] is an array of w_bit_set objects of
	/// length Lambda[i] (after \c init_arrangement_cells() has been
	/// called properly).
	w_bit_set ** C;

	/// A[i] essentialy contains the same information as C[i]. The
	/// crucial differnce is that each A[i] is an stl multiset, so that
	/// \c enumerate_differences can be implemented more
	/// efficiently. \c A is used only by \c enumerate_differences(),
	/// in all other procedures \c A is \c NULL.
	arrangement* A;

	/// P[i][j] contains the position of the cell C[i][j] in T[i].
	int ** P;
 
	/// M[i][j][nu] contains the edit distance with that the nu-th
	/// occurence of s_j occurs in T_i. This value is the basis for
	/// computing the weight of a simultaneous occurence.
	int *** M;
	
	int L;

	/// This variable contains the result of the optimization
	/// procedure.
	float overall_best_weight;

	/// h_best_weight contains the h-th best solution, in case the -h
	/// option was specified.
	float h_best_weight;

	/// best_nu[i] is the starting position of the best solution's
	/// interval in T[i].
	int* best_nu;

	/// best_nu[i] is the starting position of a temporary solution in
	/// the course of algorithm (A2), as implemented in \c
	/// enumerate_subsets() and its subprocedures.
	int* last_nu;

	/// Variable for h-best-solutions algorithm.  best_h_nu[eta][i] is
	/// the starting position of the eta-th best solution's interval in
	/// T[i].
	int** best_h_nu;
	
	/// Variable for h-best-solutions algorithm. the eta-th best solution
	/// has weight best_h_weight[eta].
	float* best_h_wghts;

	/// Variable for h-best-solutions algorithm. the eta-th best solution
	/// has weight best_h_weight[eta].
	w_bit_set* best_h_intersections;

	/// Number of solutions found so far.
	int num_of_solutions;

	/// The subset of [1:k] representing the best solution found so
	/// far.
	w_bit_set best_intersection;

	/// Maximize weighted sets instead of set cardinality
	bool weighted;

	/// Different weighting schemes can be applied:
	///
	/// 0 : corresponds to unweighted
	/// 1 : a weight is computed based on the number of mismatches of a binding site occurence
	/// 2 : an advanced weighting scheme based on logs of p-value-like measures is applied.
	int weight_type;

	/// Take into account matches of the reverse complement of each
	/// s[j] (i.e., occurences on both the 3' and the 5' strand of the
	/// genome sequences)
	bool reverse;

	/// print which proportion of the search space has been
	/// examined. Due to the branch-and-bound-approach, this proportion
	/// is NOT immediately correlated to the remaining time. However,
	/// it gives a basic idea of how much work remains to be done when
	/// dealing with larger instances.
	bool report_progress;

	/// for converting C,G,T,A to 0,1,2,3
	int* index_table;

	/// frequencies[i][x] contains the absolute frequency of nucleotide
	/// x in T[i], for x=C,G,T,A.
	int** frequencies;

	/// frequencies[i][x][y] contains the absolute frequency of the
	/// dinucleotide sequence xy in T[i], for x,y=C,G,T,A.
	int*** di_frequencies;
	
public:

	/// default constructor.
	weighted_footprint_detector();

	weighted_footprint_detector(int,int,char**,char**,char**,char**,int*,int*,int*);

	/// destructor.
	~weighted_footprint_detector();

	/// free all memory used.
	void clear();

	/// based on the occurences of the fragments s in the genomes T,
	/// compute the cells C[i] of the interval arrangement induced by
	/// the occurences in each T[i].
	void init_arrangement_cells();

	/// bound function for branch-and-bound procedure enumerate_subsets
	/// (Algorithm (A2))
	bool bound(const w_bit_set& M, const unsigned int& j);
	
	/// bound function for the weighted version of the branch-and-bound
	/// procedure enumerate_subsets (Algorithm (A2))
	bool w_bound(const w_bit_set& M, const unsigned int& j);

	/// bound function for the h-best-match and weighted version of the
	/// branch-and-bound procedure enumerate_subsets (Algorithm
	/// (A2)). Uses a difference policy in the same way as \c
	/// compute_h_best_intersections().
	bool h_best_w_bound(const w_bit_set& M, const unsigned int& j);

	/// needed for the weighted version of the branch-and-bound-version
	/// of Algorithm (A1) implemented in \c enumerate_subsets()
	void enum_w_sets(w_bit_set M, int j);

	/// weighted version for computing the h best solutions using
	/// Algorithm (A2). Uses a difference policy in the same way as \c
	/// compute_h_best_intersections().
	void enum_h_best_w_sets(w_bit_set M, int j);

	/// unweighted version of computing the h best solutions using
	/// Algorithm (A2). Uses a difference policy in the same way as \c
	/// compute_h_best_intersections().
	void enum_h_best_sets(w_bit_set M, int j);
	
	/// needed for the branch-and-bound-version of Algorithm (A1)
	/// implemented in \c enumerate_subsets()
	void enum_sets(w_bit_set M, int j);

	/// implements the branch-and-bound-version of Algorithm (A2)
	void enumerate_subsets();

	/// implements the branch-and-bound-version of Algorithm (A2) for
	/// finding the \c h best matches.
	void enumerate_h_best_subsets();
	
	/// implements the branch-and-bound-version of Algorithm (A1)
	void compute_intersections();

	/// if a new hit was found that might belong to the list of the h
	/// best solutions, this procedure is called by \c
	/// compute_h_best_intersections(). If the new candidate is
	/// sufficiently different from the h best solutions found so far,
	/// the list of the h best solutions is updated by the new solution
	/// correspondingly.
	void update_h_best_list(const float& wght, int* nu, w_bit_set& intersection);

	/// compute the h best hits using the branch-and-bound-version of
	/// algorithm (A1)
	void compute_h_best_intersections();

	/// return \c M setminus \c D
	w_bit_set difference(w_bit_set M, w_bit_set& D);
	
	/// test whether M setminus D is a suitable barbque.
	bool test_diff_set(w_bit_set& M);

	/// return l=M.cardinality(), with E[0],...E[l-1] containing the l
	/// elements contained in M.
	void get_element_array(w_bit_set M, int* E, int& l);
	
	/// compute a best barbeque with delta-bounded difference based on
	/// a parameterized version of the bbq problem.
	void enumerate_differences(int delta);
	
	/// compute the occurences of s[j] in T[i], and produce \c
	/// match_list as a result.
	void get_frag_occurences(int i, int j, match_list&);

	/// compute the occurences of s[j] in T[i]	
	void get_occurences(int i, int j);

	/// compute the occurences of the fragment in group j in T[i]
	void get_group_occurences(int i, int j);

	/// start the matching process
	void run(int,float);

	/// add the reverted sequence fragments to the set of all
	/// fragments, so that occurences along both the 3' strand and the
	/// 5' strand are taken into account.
	void reverse_compl();

	/// return the best solution found so far.
	float get_result();

	/// print the best h solution found so far into stdout.
	void print_h_results();

	/// print the solution specified by \c intersection into stdout.
	float print_single_result(int* nu, w_bit_set& intersection, const float& weight);

	/// print the result of the whole alignment procedure into stdout.
	float print_result();

	/// print the result of the whole alignment procedure into a
	/// postscript file named \c filename.
	void print_postscript(std::string filename);

	/// print labels of the current instance into a postscript file
	/// named \c filename.
	void print_ps_labels(std::string filename);
	
	/// print the best h results of the whole alignment procedure into
	/// a postscript file named \c filename.
	void print_ps_h_results(std::string&);

	/// print the solution specified by \c nu into a postscript file
	/// associated with \c o
	void print_ps_single_result(int* nu, std::ostream& o, w_bit_set& intersection);
	
	//int intersection_cardinality(unsigned int);

	/// print the names of the sequences (or groups) contained in the
	/// argument passed.
	void print_with_names(w_bit_set&);

	/// set the number h of hits to be computed, i.e., the best h
	/// (substantially different) solutions are computed.
	void set_number_of_hits(int);

	/// switch weighted intersections on and off.
	void set_weighted(bool);

	/// switch between couting and printing results for thresholded
	/// instances.
	void set_count_results(bool);

	/// choose a weighting scheme. See \c ::weight_type for weighting
	/// schemes available.
	void set_weight_type(int wt);

	/// if set to true, the fragments \c s[k/2],...,s[k-1] are regarded
	/// as fragments on the 5' strand, while \c s[1],...,s[k/2-1] are
	/// regarded as fragments on the 3' strand.
	void set_reverse(bool);

	/// switch weighted intersections on and off.
	void set_postscript(bool);

	/// switch weighted intersections on and off.
	void set_ps_filename(std::string);

	/// switch printing of  progress report during execution.
	void set_progress_report(bool);
	
	/// Choose the algorithm to be applied. A=1 chooses (A1) (default),
	/// A=2 chooses (A2). All other values are ignored and leave the
	/// current method unchanged.
	void set_algorithm(int A);

	/// switch grouping of fragments on and off.
	void set_grouping(bool);

	/// switch number of leading chars used for grouping of fragments
	/// on and off.
	void set_group_characters(int);
	
	/// compute group names form fragment names as well as the
	/// corresponding mapping group_indices that assigns a set of fragment
	/// index to each group index.
	void get_groups();

	/// pass true if more detailled output is desired.
	void set_output_detail(bool);

	/// set the bound for \c enumerate_differences(). If \c set_delta()
	/// is called, matching will automatically be performed by 
	/// \c enumerate_differences().
	void set_delta(int);
	
	//void get_grouped_occurences(int i, int j);

	/// increase the index vector nu at position I. If nu[I]>Lambda[I],
	/// then an overflow is handled, i.e., nu[I+1] is increased. This
	/// procedure is needed by \c compute_intersections() for the
	/// branch-and-bound version of Algorithm (A1).
	bool increase(int* nu,int I);

	/// When calling compute_h_best_intersections(), we want to obtain
	/// substantially different solutions. The procedure \c differs()
	/// decides whether the two solutions given as an argument are
	/// "substantially different". In fact, there are numerous ways of
	/// formally defining solutions to be different. Hence, this
	/// procedure may get outsorced in a template class in later
	/// versions, so that several difference policies can be
	/// implemented.
	bool differs(int*,int*);

	/// Compute a first approximation to a p-value like measure of
	/// significance for the nu-th occurence of s[j] in T[i]. The log
	/// of this p-value is used for an advanced weighting scheme.
	double p_value(int i, int j, int nu);

	/// We count neucleotide frequencies and dinucleotide frequencies
	/// in the genome, so that we can determine (approximations of) p-
	/// and E-values later on.
	void get_freq();

	/// compute a weight for the nu-th occurence of \c s[j] in \c T[i]
	/// (with position \c p[i][j][nu] )
	float get_weight(int i, int j, int nu);
		
	/// do some preprocessing for \c p_value() and get_freq()
	void compute_index_table();
};

template class weighted_footprint_detector<weighted_variable_bitset>;
template class weighted_footprint_detector<weighted_bit_multi_set>;
template class weighted_footprint_detector<bit_multi_set>;

typedef weighted_footprint_detector<weighted_variable_bitset> w_set_footprint_detector;
typedef weighted_footprint_detector<bit_multi_set> w_mset_footprint_detector;
typedef weighted_footprint_detector<weighted_bit_multi_set> w_wmset_footprint_detector;

/// template instantiation

} // end namespace bbq


#include "weighted_footprint_detector.cpp"

#endif


